/*
 * Copyright 2016  Samsung Electronics Co., Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dlog.h>
#include <gio/gio.h>
#include <sys/smack.h>
#include <badge.h>
#include <badge_db.h>
#include <badge_internal.h>
#include <badge_setting.h>
#include <badge_setting_service.h>
#include <badge_ipc.h>
#include <tzplatform_config.h>

#include "service_common.h"
#include "badge_service.h"
#include "debug.h"

#define PROVIDER_BADGE_INTERFACE_NAME "org.tizen.data_provider_badge_service"
static GHashTable *_monitoring_hash;

static void _on_name_appeared(GDBusConnection *connection,
		const gchar     *name,
		const gchar     *name_owner,
		gpointer         user_data)
{
	DBG("name [%s]", name);
}

static void _on_name_vanished(GDBusConnection *connection,
		const gchar     *name,
		gpointer         user_data)
{
	DBG("name [%s]", name);
	monitoring_info_s *info = (monitoring_info_s *)user_data;

	if (info) {
		DBG("vanished uid [%d]", info->uid);
		g_bus_unwatch_name(info->watcher_id);
		delete_monitoring_list(&_monitoring_hash, name, info->uid);

		if (info->bus_name)
			free(info->bus_name);
		free(info);
	}
}

static void _badge_dbus_method_call_handler(GDBusConnection *conn,
		const gchar *sender, const gchar *object_path,
		const gchar *iface_name, const gchar *method_name,
		GVariant *parameters, GDBusMethodInvocation *invocation,
		gpointer user_data)
{
	DBG("badge method_name [%s]", method_name);

	GVariant *reply_body = NULL;
	int ret = BADGE_ERROR_INVALID_PARAMETER;
	uid_t uid = get_sender_uid(sender);
	pid_t pid = get_sender_pid(sender);

	if (g_strcmp0(method_name, "badge_service_register") == 0)
		ret = service_register(parameters, &reply_body, sender,
				_on_name_appeared, _on_name_vanished, &_monitoring_hash, uid);
	else if (g_strcmp0(method_name, "get_badge_existing") == 0)
		ret = badge_get_badge_existing(parameters, &reply_body, uid);
	else if (g_strcmp0(method_name, "get_list") == 0)
		ret = badge_get_badge_list(parameters, &reply_body, uid);
	else if (g_strcmp0(method_name, "insert_badge") == 0)
		ret = badge_insert(parameters, &reply_body, uid);
	else if (g_strcmp0(method_name, "delete_badge") == 0)
		ret = badge_delete(parameters, &reply_body, uid, pid);
	else if (g_strcmp0(method_name, "set_badge_count") == 0)
		ret = badge_set_badge_count(parameters, &reply_body, uid, pid);
	else if (g_strcmp0(method_name, "get_badge_count") == 0)
		ret = badge_get_badge_count(parameters, &reply_body, uid);
	else if (g_strcmp0(method_name, "set_disp_option") == 0)
		ret = badge_set_display_option(parameters, &reply_body, uid);
	else if (g_strcmp0(method_name, "get_disp_option") == 0)
		ret = badge_get_display_option(parameters, &reply_body, sender, uid);
	else if (g_strcmp0(method_name, "update_badge_setting") == 0)
		ret = badge_update_badge_setting(parameters, &reply_body, uid);
	else if (g_strcmp0(method_name, "get_setting_by_appid") == 0)
		ret = badge_get_setting_by_appid(parameters, &reply_body, uid);
	else if (g_strcmp0(method_name, "init_badge") == 0)
		ret = badge_init_display(parameters, &reply_body, uid);

	if (ret == BADGE_ERROR_NONE) {
		DBG("Badge service success, method [%s]", method_name);
		g_dbus_method_invocation_return_value(
				invocation, reply_body);
	} else {
		ERR("Badge service fail, method_name[%s] err[%d]",
			method_name, ret);
		g_dbus_method_invocation_return_error(
				invocation,
				BADGE_ERROR,
				ret,
				"badge service error");
	}
}

static const GDBusInterfaceVTable _badge_interface_vtable = {
		_badge_dbus_method_call_handler,
		NULL,
		NULL
};

int badge_register_dbus_interface(void)
{
	static gchar introspection_xml[] =
			"  <node>"
			"  <interface name='org.tizen.data_provider_badge_service'>"
			"        <method name='badge_service_register'>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='get_badge_existing'>"
			"          <arg type='s' name='pkgname' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='i' name='exist' direction='out'/>"
			"        </method>"

			"        <method name='get_list'>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='a(v)' name='badge_list' direction='out'/>"
			"        </method>"

			"        <method name='insert_badge'>"
			"          <arg type='s' name='pkgname' direction='in'/>"
			"          <arg type='s' name='writable_pkg' direction='in'/>"
			"          <arg type='s' name='caller' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='delete_badge'>"
			"          <arg type='s' name='pkgname' direction='in'/>"
			"          <arg type='s' name='caller' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='set_badge_count'>"
			"          <arg type='s' name='pkgname' direction='in'/>"
			"          <arg type='s' name='caller' direction='in'/>"
			"          <arg type='i' name='count' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='get_badge_count'>"
			"          <arg type='s' name='pkgname' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='i' name='count' direction='out'/>"
			"        </method>"

			"        <method name='set_disp_option'>"
			"          <arg type='s' name='pkgname' direction='in'/>"
			"          <arg type='s' name='caller' direction='in'/>"
			"          <arg type='i' name='is_display' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='get_disp_option'>"
			"          <arg type='s' name='pkgname' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='i' name='is_display' direction='out'/>"
			"        </method>"

			"        <method name='update_badge_setting'>"
			"          <arg type='s' name='pkgname' direction='in'/>"
			"          <arg type='s' name='appid' direction='in'/>"
			"          <arg type='i' name='allow_to_display' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='get_setting_by_appid'>"
			"          <arg type='s' name='appid' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='v' name='setting' direction='out'/>"
			"        </method>"

			"        <method name='init_badge'>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"  </interface>"
			"  </node>";

	return service_common_register_dbus_interface(introspection_xml, _badge_interface_vtable);
}

static int _validate_and_set_param_uid_with_uid(uid_t uid, uid_t *param_uid)
{
	if (uid > NORMAL_UID_BASE && uid != *param_uid) {
		ERR("Invalid sender uid[%d] param_uid[%d]", uid, *param_uid);
		return BADGE_ERROR_INVALID_PARAMETER;
	} else if (uid <= NORMAL_UID_BASE) {
		if (*param_uid <= NORMAL_UID_BASE) {
			if (*param_uid != uid)
				return BADGE_ERROR_INVALID_PARAMETER;
			*param_uid = tzplatform_getuid(TZ_SYS_DEFAULT_USER);
		}
	}
	return BADGE_ERROR_NONE;
}

static void _release_badge_info(gpointer data)
{
	badge_info_s *badge;

	badge = (badge_info_s *)data;
	if (badge == NULL)
		return;
	if (badge->pkg)
		free(badge->pkg);
	free(badge);
}

/* get_badge_existing */
int badge_get_badge_existing(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret = BADGE_ERROR_NONE;
	char *pkgname = NULL;
	bool existing = 0;
	uid_t param_uid;

	g_variant_get(parameters, "(&si)", &pkgname, &param_uid);

	if (pkgname == NULL)
		return BADGE_ERROR_INVALID_PARAMETER;

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != BADGE_ERROR_NONE)
		return ret;

	ret = badge_db_is_existing(pkgname, &existing, param_uid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to get badge existing :%d", ret);
		return ret;
	}
	INFO("The badge of [%s] exists", pkgname);

	*reply_body = g_variant_new("(i)", existing);
	if (*reply_body == NULL) {
		ERR("Failed to make gvariant");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	DBG("result [%d]", ret);
	return ret;
}

/* get_list */
int badge_get_badge_list(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	GList *badge_list = NULL;
	GList *iter_list = NULL;
	GVariant *body = NULL;
	badge_info_s *badge;
	GVariantBuilder *builder;
	int ret;
	uid_t param_uid;

	g_variant_get(parameters, "(i)", &param_uid);

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != BADGE_ERROR_NONE)
		return ret;

	ret = badge_db_get_list(&badge_list, param_uid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to get list [%d]", ret);
		return ret;
	}

	builder = g_variant_builder_new(G_VARIANT_TYPE("a(v)"));
	iter_list = g_list_first(badge_list);
	for (; iter_list != NULL; iter_list = iter_list->next) {
		badge = iter_list->data;
		body = g_variant_new("(si)", badge->pkg, badge->badge_count);
		g_variant_builder_add(builder, "(v)", body);
	}
	g_list_free_full(badge_list, (GDestroyNotify)_release_badge_info);

	*reply_body = g_variant_new("(a(v))", builder);
	g_variant_builder_unref(builder);

	if (*reply_body == NULL) {
		ERR("Failed to make reply");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	DBG("Success to get badge list");
	return BADGE_ERROR_NONE;
}

/* insert_badge */
int badge_insert(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret = BADGE_ERROR_NONE;
	char *pkgname = NULL;
	char *writable_pkg = NULL;
	char *caller = NULL;
	GVariant *body = NULL;
	uid_t param_uid;
	int allow_to_display = 0;

	g_variant_get(parameters, "(&s&s&si)", &pkgname, &writable_pkg, &caller, &param_uid);

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != BADGE_ERROR_NONE)
		return ret;

	if (pkgname != NULL && writable_pkg != NULL && caller != NULL)
		ret = badge_db_insert(pkgname, writable_pkg, caller, param_uid);
	else
		return BADGE_ERROR_INVALID_PARAMETER;

	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to insert badge [%d]", ret);
		return ret;
	}

	ret = badge_db_get_allow_to_display_by_appid(pkgname, &allow_to_display, param_uid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to get allow_to_display by appid [%d]", ret);
		return ret;
	}

	if (allow_to_display) {
		body = g_variant_new("(si)", pkgname, param_uid);
		if (body == NULL) {
			ERR("Failed to make gvariant");
			return BADGE_ERROR_OUT_OF_MEMORY;
		}

		ret = send_notify(body, "insert_badge_notify", &_monitoring_hash, PROVIDER_BADGE_INTERFACE_NAME, param_uid);
		g_variant_unref(body);

		if (ret != BADGE_ERROR_NONE) {
			ERR("Failed to send notify [%d]", ret);
			return ret;
		}
	}

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make reply");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	INFO("Success, pkgname[%s] writable_pkg[%s] caller[%s]", pkgname, writable_pkg, caller);

	return ret;
}

/* delete_badge */
int badge_delete(GVariant *parameters, GVariant **reply_body, uid_t uid, pid_t pid)
{
	int ret = BADGE_ERROR_NONE;
	char *pkgname = NULL;
	char *caller = NULL;
	GVariant *body = NULL;
	uid_t param_uid;

	g_variant_get(parameters, "(&s&si)", &pkgname, &caller, &param_uid);

	if (pkgname == NULL || caller == NULL)
		return BADGE_ERROR_INVALID_PARAMETER;

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != BADGE_ERROR_NONE)
		return ret;

	ret = badge_db_delete(pkgname, caller, param_uid, pid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to delete badge [%d]", ret);
		return ret;
	}

	body = g_variant_new("(si)", pkgname, param_uid);
	if (body == NULL) {
		ERR("Failed to make gvariant");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	ret = send_notify(body, "delete_badge_notify", &_monitoring_hash, PROVIDER_BADGE_INTERFACE_NAME, param_uid);
	g_variant_unref(body);

	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to send notify:%d", ret);
		return ret;
	}

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make reply");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	INFO("success, pkgname[%s] caller[%s]", pkgname, caller);

	return ret;
}

/* set_badge_count */
int badge_set_badge_count(GVariant *parameters, GVariant **reply_body, uid_t uid, pid_t pid)
{
	int ret = BADGE_ERROR_NONE;
	char *pkgname = NULL;
	char *caller = NULL;
	unsigned int count = 0;
	GVariant *body = NULL;
	uid_t param_uid;
	int allow_to_display = 0;

	g_variant_get(parameters, "(&s&sii)", &pkgname, &caller, &count, &param_uid);

	if (pkgname == NULL || caller == NULL)
		return BADGE_ERROR_INVALID_PARAMETER;

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != BADGE_ERROR_NONE)
		return ret;

	ret = badge_db_set_count(pkgname, caller, count, param_uid, pid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to set badge [%d]", ret);
		return ret;
	}

	ret = badge_db_get_allow_to_display_by_appid(pkgname, &allow_to_display, param_uid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to get allow_to_display [%d]", ret);
		return ret;
	}

	if (allow_to_display) {
		body = g_variant_new("(sii)", pkgname, count, param_uid);
		if (body == NULL) {
			ERR("Failed to make gvariant");
			return BADGE_ERROR_OUT_OF_MEMORY;
		}

		ret = send_notify(body, "set_badge_count_notify", &_monitoring_hash, PROVIDER_BADGE_INTERFACE_NAME, param_uid);
		g_variant_unref(body);

		if (ret != BADGE_ERROR_NONE) {
			ERR("Failed to send notify [%d]", ret);
			return ret;
		}
	}

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make reply");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	INFO("Success, pkgname[%s] caller[%s] count[%d]", pkgname, caller, count);

	return ret;
}

/* get_badge_count */
int badge_get_badge_count(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret = BADGE_ERROR_NONE;
	char *pkgname = NULL;
	unsigned int count = 0;
	uid_t param_uid;

	g_variant_get(parameters, "(&si)", &pkgname, &param_uid);

	if (pkgname == NULL)
		return BADGE_ERROR_INVALID_PARAMETER;

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != BADGE_ERROR_NONE)
		return ret;

	ret =  badge_db_get_count(pkgname, &count, param_uid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to get badge count [%d]", ret);
		return ret;
	}

	*reply_body = g_variant_new("(i)", count);
	if (*reply_body == NULL) {
		ERR("Failed to make gvariant");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	INFO("Success, count[%d]", count);
	return ret;
}

/* set_disp_option */
int badge_set_display_option(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret = BADGE_ERROR_NONE;
	char *pkgname = NULL;
	char *caller = NULL;
	unsigned int is_display = 0;
	GVariant *body = NULL;
	uid_t param_uid;
	int allow_to_display = 0;

	g_variant_get(parameters, "(&s&sii)", &pkgname, &caller, &is_display, &param_uid);

	if (pkgname == NULL || caller == NULL)
		return BADGE_ERROR_INVALID_PARAMETER;

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != BADGE_ERROR_NONE)
		return ret;

	ret = badge_db_set_display_option(pkgname, is_display, param_uid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to set display option [%d]", ret);
		return ret;
	}

	ret = badge_db_get_allow_to_display_by_appid(pkgname, &allow_to_display, param_uid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to get allow_to_display [%d]", ret);
		return ret;
	}

	if (allow_to_display) {
		body = g_variant_new("(sii)", pkgname, is_display, param_uid);
		if (body == NULL) {
			ERR("Failed to make gvariant");
			return BADGE_ERROR_OUT_OF_MEMORY;
		}

		ret = send_notify(body, "set_disp_option_notify", &_monitoring_hash, PROVIDER_BADGE_INTERFACE_NAME, param_uid);
		g_variant_unref(body);

		if (ret != BADGE_ERROR_NONE) {
			ERR("Failed to send notify [%d]", ret);
			return ret;
		}
	}

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make reply");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	INFO("Success, pkgname[%s] caller[%s] is_dispaly[%d]",
		pkgname, caller, is_display);

	return ret;
}

static gint __compare_sender(gconstpointer a, gconstpointer b)
{
	if (!a)
		return -1;

	return strcmp(a, b);
}

/* get_disp_option */
int badge_get_display_option(GVariant *parameters, GVariant **reply_body, const gchar *sender, uid_t uid)
{
	int ret = BADGE_ERROR_NONE;
	char *pkgname = NULL;
	unsigned int is_display = 0;
	int allow_to_display;
	GList *monitoring_list = NULL;
	GList *find_list = NULL;
	uid_t param_uid;

	g_variant_get(parameters, "(&si)", &pkgname, &param_uid);

	if (pkgname == NULL)
		return BADGE_ERROR_INVALID_PARAMETER;

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != BADGE_ERROR_NONE)
		return ret;

	ret = badge_db_get_display_option(pkgname, &is_display, param_uid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to get display option [%d]", ret);
		return ret;
	}

	ret = badge_db_get_allow_to_display_by_appid(pkgname, &allow_to_display, param_uid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to get allow_to_display [%d]", ret);
		return ret;
	}

	if (!allow_to_display) {
		monitoring_list = (GList *)g_hash_table_lookup(_monitoring_hash, GUINT_TO_POINTER(param_uid));
		monitoring_list = g_list_first(monitoring_list);
		find_list = g_list_find_custom(monitoring_list, (gconstpointer)sender,
					       (GCompareFunc)__compare_sender);
		if (find_list != NULL)
			is_display = 0;
	}

	*reply_body = g_variant_new("(i)", is_display);
	if (*reply_body == NULL) {
		ERR("Failed to make reply");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	INFO("Success, display option[%d]", is_display);
	return ret;
}

static int __badge_emit_fake_signal(char *appid, unsigned int disp_option, uid_t uid)
{
	int ret;
	GVariant *fake_body = NULL;

	fake_body = g_variant_new("(sii)", appid, disp_option, uid);
	if (fake_body == NULL) {
		ERR("Failed to make gvariant");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	ret = send_notify(fake_body, "set_disp_option_notify", &_monitoring_hash, PROVIDER_BADGE_INTERFACE_NAME, uid);
	g_variant_unref(fake_body);
	if (ret != BADGE_ERROR_NONE)
		ERR("Failed to send notify [%d]", ret);

	return ret;
}

/* update_badge_setting */
int badge_update_badge_setting(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret = BADGE_ERROR_NONE;
	uid_t param_uid;
	unsigned int allow_to_display;
	unsigned int disp_option = 0;
	char *pkgname = NULL;
	char *appid = NULL;

	g_variant_get(parameters, "(&s&sii)", &pkgname, &appid, &allow_to_display, &param_uid);

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != BADGE_ERROR_NONE)
		return ret;

	ret = badge_db_update_setting(pkgname, appid, allow_to_display, param_uid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to update badge setting [%d]", ret);
		return ret;
	}

	if (!allow_to_display) {
		ret = __badge_emit_fake_signal(appid, allow_to_display, param_uid);
		if (ret != BADGE_ERROR_NONE)
			return ret;
	} else {
		ret = badge_db_get_display_option(appid, &disp_option, param_uid);
		if (ret != BADGE_ERROR_NONE && ret != BADGE_ERROR_NOT_EXIST) {
			ERR("Failed to get display_option [%d]", ret);
			return ret;
		}

		if (disp_option) {
			ret = __badge_emit_fake_signal(appid, disp_option, param_uid);
			if (ret != BADGE_ERROR_NONE)
				return ret;
		}
	}

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	INFO("Success, pkgname[%s], appid[%s]", pkgname, appid);
	return ret;
}

int badge_get_setting_by_appid(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	char *appid = NULL;
	GVariant *body;
	badge_setting_h setting = NULL;
	uid_t param_uid;

	g_variant_get(parameters, "(&si)", &appid, &param_uid);

	if (appid == NULL)
		return BADGE_ERROR_INVALID_PARAMETER;

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = badge_db_get_setting_by_appid(appid, &setting, param_uid);
	if (ret == BADGE_ERROR_NONE) {
		body = badge_ipc_make_gvariant_from_setting(setting);
		badge_setting_free_setting(setting);
		if (body == NULL) {
			ERR("Failed to make gvariant from setting");
			return NOTIFICATION_ERROR_OUT_OF_MEMORY;
		}
	} else {
		return ret;
	}

	*reply_body = g_variant_new("(v)", body);
	if (*reply_body == NULL) {
		ERR("Failed to make reply");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	INFO("Success, appid[%s]", appid);
	return ret;
}

/* init badge */
int badge_init_display(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret = BADGE_ERROR_NONE;
	unsigned int disp_option;
	int allow_to_display;
	GList *badge_list = NULL;
	GList *iter_list  = NULL;
	GVariant *body = NULL;
	badge_info_s *badge;
	uid_t param_uid;

	g_variant_get(parameters, "(i)", &param_uid);

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != BADGE_ERROR_NONE)
		return ret;

	ret = badge_db_get_list(&badge_list, param_uid);
	if (ret != BADGE_ERROR_NONE) {
		ERR("Failed to get list [%d]", ret);
		return ret;
	}

	iter_list = g_list_first(badge_list);
	while (iter_list) {
		badge = (badge_info_s *)g_list_nth_data(iter_list, 0);
		iter_list = g_list_next(iter_list);

		ret = badge_db_get_display_option(badge->pkg, &disp_option, param_uid);
		if (ret != BADGE_ERROR_NONE) {
			ERR("Failed to get display option [%d]", ret);
			continue;
		}

		ret = badge_db_get_allow_to_display_by_appid(badge->pkg, &allow_to_display, param_uid);
		if (ret != BADGE_ERROR_NONE) {
			ERR("Failed to get allow_to_display [%d]", ret);
			continue;
		}

		if (disp_option && allow_to_display) {
			body = g_variant_new("(sii)", badge->pkg, disp_option, param_uid);
			ret = send_notify(body, "set_disp_option_notify", &_monitoring_hash, PROVIDER_BADGE_INTERFACE_NAME, param_uid);
			if (ret != BADGE_ERROR_NONE) {
				ERR("Failed to send notify");
				continue;
			}
			g_variant_unref(body);
		}
	}

	if (badge_list)
		g_list_free_full(badge_list, (GDestroyNotify)_release_badge_info);

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	DBG("result [%d]", ret);
	return ret;
}

static gboolean __refresh_setting_table(gpointer data)
{
	uid_t uid = tzplatform_getuid(TZ_SYS_DEFAULT_USER);

	badge_setting_refresh_setting_table(uid);
	badge_setting_arrange_tables(uid);

	return G_SOURCE_REMOVE;
}

/*!
 * MAIN THREAD
 * Do not try to do anyother operation in these functions
 */
HAPI int badge_service_init(void)
{
	int result;

	_monitoring_hash = g_hash_table_new(g_direct_hash, g_direct_equal);
	result = badge_db_init();
	if (result != BADGE_ERROR_NONE) {
		ERR("Failed to init DB [%d]", result);
		return result;
	}

	result = badge_register_dbus_interface();
	if (result != SERVICE_COMMON_ERROR_NONE) {
		ERR("Failed to register dbus interface [%d]", result);
		return BADGE_ERROR_IO_ERROR;
	}

	g_idle_add(__refresh_setting_table, NULL);
	INFO("Successfully initialized");
	return BADGE_ERROR_NONE;
}

HAPI int badge_service_fini(void)
{
	return BADGE_ERROR_NONE;
}

/* End of a file */
