/*
 * Copyright 2016  Samsung Electronics Co., Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/signalfd.h>
#include <signal.h>
#include <ctype.h>

#include <systemd/sd-daemon.h>
#include <glib.h>
#include <glib-object.h>
#include <aul.h>
#include <vconf.h>
#include <dlog.h>
#include <locale.h>

#include "debug.h"
#include "service_common.h"
#include "shortcut_service.h"
#include "notification_service.h"
#include "notification_ex_service.h"
#include "badge_service.h"

static GMainLoop *main_loop;
static GIOChannel *channel;
static guint source;

static void lang_key_changed_cb(keynode_t *node, void *first)
{
	char *lang;
	char *r;

	lang = vconf_get_str(VCONFKEY_LANGSET);
	if (lang) {
		setenv("LANG", lang, 1);
		setenv("LC_MESSAGES", lang, 1);
		r = setlocale(LC_ALL, "");
		if (r == NULL) {
			r = setlocale(LC_ALL, lang);
			if (r != NULL)
				DBG("setlocale = %s", r);
		}
		DBG("setlocale [%s]", r);
		free(lang);
	}
}

static inline int app_create(int restart_count)
{
	int ret;

	ret = vconf_notify_key_changed(VCONFKEY_LANGSET, lang_key_changed_cb, NULL);
	if (ret < 0)
		DBG("VCONFKEY_LANGSET notify key chenaged [%d]", ret);

	lang_key_changed_cb(NULL, NULL);

	ret = notification_ex_service_init(restart_count);
	if (ret < 0)
		WARN("notification_ex [%d]", ret);

	service_common_set_connection(
		notification_ex_service_get_gdbus_connection());

	ret = shortcut_service_init();
	if (ret < 0)
		WARN("shortcut [%d]", ret);

	ret = notification_service_init(restart_count);
	if (ret < 0)
		WARN("notification [%d]", ret);

	ret = badge_service_init();
	if (ret < 0)
		WARN("badge [%d]", ret);

	service_common_init();
	INFO("Successfully initialized");

	return 0;
}

static inline int app_terminate(void)
{
	int ret;

	ret = badge_service_fini();
	if (ret < 0)
		DBG("badge [%d]", ret);

	ret = notification_service_fini();
	if (ret < 0)
		DBG("notification [%d]", ret);

	ret = shortcut_service_fini();
	if (ret < 0)
		DBG("shortcut [%d]", ret);

	return 0;
}

static gboolean __signal_handler(GIOChannel *channel,
		GIOCondition cond, gpointer data)
{
	struct signalfd_siginfo fdsi;
	ssize_t size;
	int sfd;

	sfd = g_io_channel_unix_get_fd(channel);

	size = read(sfd, &fdsi, sizeof(struct signalfd_siginfo));
	if (size != sizeof(struct signalfd_siginfo)) {
		ERR("signal read failed [%d]", errno);
		return TRUE;
	}

	if (fdsi.ssi_signo == SIGTERM) {
		ERR("Terminated(SIGTERM)");
		g_main_loop_quit(main_loop);
	} else {
		ERR("Unknown SIG[%d] received", fdsi.ssi_signo);
	}

	return TRUE;
}

static int __init_signal_handler(void)
{
	sigset_t mask;
	int sfd;
	int ret;

	ret = sigemptyset(&mask);
	if (ret < 0) {
		ERR("sigemptyset : %d", errno);
		return -1;
	}

	ret = sigaddset(&mask, SIGTERM);
	if (ret < 0) {
		ERR("sigaddset: %d", errno);
		return -1;
	}

	ret = sigprocmask(SIG_BLOCK, &mask, NULL);
	if (ret < 0) {
		ERR("sigprocmask: %d", errno);
		return -1;
	}

	sfd = signalfd(-1, &mask, SFD_NONBLOCK);
	if (sfd < 0) {
		ERR("signalfd: %d", errno);
		return -1;
	}

	channel = g_io_channel_unix_new(sfd);
	g_io_channel_set_close_on_unref(channel, TRUE);
	g_io_channel_set_encoding(channel, NULL, NULL);
	g_io_channel_set_buffered(channel, FALSE);

	source = g_io_add_watch(channel, G_IO_IN, __signal_handler, NULL);

	INFO("source[%u]", source);

	return 0;
}

void __finish(void)
{
	if (source > 0)
		g_source_remove(source);
	if (channel)
		g_io_channel_unref(channel);

	app_terminate();
}

int main(int argc, char *argv[])
{
	int restart_count = 0;

	main_loop = g_main_loop_new(NULL, FALSE);
	if (!main_loop)
		return -1;

#if (GLIB_MAJOR_VERSION <= 2 && GLIB_MINOR_VERSION < 36)
	g_type_init();
#endif

	vconf_get_int(VCONFKEY_MASTER_RESTART_COUNT, &restart_count);

	app_create(restart_count);
	sd_notify(0, "READY=1");

	restart_count++;
	vconf_set_int(VCONFKEY_MASTER_RESTART_COUNT, restart_count);

	if (__init_signal_handler() < 0)
		ERR("Failed to init signal handler");

	g_main_loop_run(main_loop);

	__finish();

	g_main_loop_unref(main_loop);

	return 0;
}

/* End of a file */
