/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dlog.h>
#include <stdio.h>
#include <stdlib.h>
#include <notification_type.h>
#include <notification_viewer.h>

#include <iostream>
#include <sstream>
#include <algorithm>
#include <map>

#include <notification-ex/manager.h>
#include <notification-ex/reporter.h>
#include <notification-ex/dbus_sender.h>
#include <notification-ex/dbus_event_listener.h>
#include <notification-ex/dbus_connection_manager.h>
#include <notification-ex/db_manager.h>
#include <notification-ex/common.h>
#include <notification-ex/exception.h>
#include <notification-ex/item_info_internal.h>
#include <notification-ex/button_item.h>
#include <notification-ex/ex_util.h>
#include <notification-ex/ievent_info_internal.h>
#include <notification-ex/iitem_info_internal.h>

#include "debug.h"
#include "notification_ex_service.h"

using namespace std;
using namespace notification;
using namespace notification::item;

class DPMFacade {
 public:
  DPMFacade(unique_ptr<Reporter> reporter, unique_ptr<Manager> manager, int restartCount)
      : reporter_(move(reporter)), manager_(move(manager)) {
    DBManager::InitializeDB();

    if (restartCount == 0)
      DBManager::InitializeData();

    hide_map_ = DBManager::GetHideMap();
  }

  void DelegateReporterEvent(const IEventInfo& info, shared_ptr<item::AbstractItem> item) {
    reporter_->SendEvent(info, item);
  }

  void DelegateReporterEvent(const IEventInfo& info,
        std::list<std::shared_ptr<item::AbstractItem>> itemList) {
    reporter_->SendEvent(info, itemList);
  }

  void DelegateManagerEvent(const IEventInfo& info, shared_ptr<item::AbstractItem> noti) {
    manager_->SendEvent(info, noti);
  }

  void LaunchDefaultViewer(list<shared_ptr<item::AbstractItem>> item,
                                   notification_op_type_e status) {
    int64_t priv_id;
    if (item.size() == 0)
      return;

    if (item.size() > 1)
      priv_id = NOTIFICATION_PRIV_ID_NONE;
    else
      priv_id = static_pointer_cast<IItemInfoInternal>(
                    item.front()->GetInfo())->GetPrivateId();

    LaunchDefaultViewer(priv_id, status,
                        static_pointer_cast<IItemInfoInternal>(
                            item.front()->GetInfo())->GetUid());
  }
  void LaunchDefaultViewer(shared_ptr<item::AbstractItem> item,
              notification_op_type_e status) {
    LaunchDefaultViewer(
        static_pointer_cast<IItemInfoInternal>(item->GetInfo())->GetPrivateId(),
        status,
        static_pointer_cast<IItemInfoInternal>(item->GetInfo())->GetUid());
  }

  void LaunchDefaultViewer(int64_t privId, notification_op_type_e status,
                                   uid_t uid) {
    int ret = notification_launch_default_viewer(
                   static_cast<int>(privId), status, uid);
    if (ret != ERROR_NONE)
      ERR("Failed to launch default viewer [%d]", ret);
  }

  unique_ptr<Reporter> reporter_;
  unique_ptr<Manager> manager_;
  map<string, string> hide_map_;
};

static DPMFacade* facade_;

class DPMReporter : public Reporter {
 protected:
  void OnUpdate(const IEventInfo& info,
      shared_ptr<item::AbstractItem> updatedItem) {
    DBG("Update !!!");
    int ret;

    /* check Manager::Hide() */
    ret = UpdateHideApp(updatedItem);
    if (ret == ERROR_NONE)
      return;

    if (ret != ERROR_NOT_EXIST_ID) {
      SendError(info, static_cast<NotificationError>(ret));
      return;
    }

    ret = DBManager::UpdateNotification(updatedItem);
    if (ret == ERROR_NONE) {
      /* noti owner */
      facade_->DelegateManagerEvent(info, updatedItem);

      /* noti to viewers */
      facade_->DelegateReporterEvent(info, updatedItem);
    } else {
      SendError(info, static_cast<NotificationError>(ret));
    }
  }

  void DoDelete(const IEventInfo& info,
      shared_ptr<item::AbstractItem> deletedItem) {
    int ret = DBManager::DeleteNotification(deletedItem);
    if (ret == ERROR_NONE) {
      /* noti owner */
      facade_->DelegateManagerEvent(info, deletedItem);

      /* noti to viewers */
      facade_->DelegateReporterEvent(info, deletedItem);
    } else {
      SendError(info, static_cast<NotificationError>(ret));
    }
  }

  void OnDelete(const IEventInfo& info,
      shared_ptr<item::AbstractItem> deletedItem) {
    DBG("Delete !!!");
    if (info.GetEventType() == IEventInfo::EventType::DeleteAll) {
      list<shared_ptr<item::AbstractItem>> noti_list =
          DBManager::GetNotificationList(((const IEventInfoInternal&)info).GetUid());
      for (auto& i : noti_list) {
        DoDelete(info, i);
      }
    } else {
      DoDelete(info, deletedItem);
    }
  }

  void OnEvent(const IEventInfo& info,
      list<shared_ptr<item::AbstractItem>> noti_list) override {
    LOGI("Event received (%d) !!", (int)info.GetEventType());
    int type = info.GetEventType();
    if (type == IEventInfo::EventType::DeleteAll) {
      OnDelete(info, nullptr);
      return;
    }

    for (auto& i : noti_list) {
      switch (type) {
      case IEventInfo::EventType::Update:
        OnUpdate(info, i);
        break;
      case IEventInfo::EventType::Delete:
        OnDelete(info, i);
        break;
      default :
        break;
      }
    }
  }

  list<shared_ptr<item::AbstractItem>> OnRequestEvent(const IEventInfo& info) override {
    DBG("Get report !!! %s", info.GetItemId().c_str());
    return DBManager::GetNotificationList(
        ((const IEventInfoInternal&)info).GetUid(),
        ((const IEventInfoInternal&)info).GetChannel().c_str());
  }

  int OnRequestNumber(const IEventInfo& info) override {
    DBG("Get count !!! %s", info.GetItemId().c_str());
    return DBManager::GetNotificationList(
        ((const IEventInfoInternal&)info).GetUid()).size();
  }

  int UpdateHideApp(shared_ptr<item::AbstractItem> updatedItem) {
    int ret;
    string hide_list;
    int uid = static_pointer_cast<IItemInfoInternal>(updatedItem->GetInfo())->GetUid();
    DBG("try update hide app_id(%s) sender(%s) uid(%s)",
      updatedItem->GetId().c_str(),
      updatedItem->GetSenderAppId().c_str(), string(to_string(uid)).c_str());
    string map_key = updatedItem->GetId() + updatedItem->GetSenderAppId() + string(to_string(uid));
    list<string> updated_hide_list = static_pointer_cast<IItemInfoInternal>(updatedItem->GetInfo())->GetHideViewerList();

    /* Check new hide app */
    auto noti = facade_->hide_map_.find(map_key);
    if (noti == facade_->hide_map_.end()) {
      if (updated_hide_list.size() > 0) {
        for (auto& i : updated_hide_list)
          hide_list += i + ";";

        ret = DBManager::UpdateHideList(updatedItem, hide_list);
        if (ret != ERROR_NONE)
          return ret;

        facade_->hide_map_[map_key] = hide_list;
        return ERROR_NONE;
      } else {
        DBG("Empty hide viewer list");
        return ERROR_NOT_EXIST_ID;
      }
    }

    /* Check new hide app */
    hide_list = noti->second;
    for (auto& i : updated_hide_list) {
      if (hide_list.find(i + ";") == string::npos) {
        string new_hide_list = hide_list + i + ";";
        ret = DBManager::UpdateHideList(updatedItem, new_hide_list);
        if (ret != ERROR_NONE)
          return ret;

        facade_->hide_map_[map_key] = new_hide_list;
        return ERROR_NONE;
      }
    }

    return ERROR_NOT_EXIST_ID;
  }

 public:
  DPMReporter(std::unique_ptr<IEventSender> sender,
      std::unique_ptr<IEventListener> listener)
    : Reporter(move(sender), move(listener)) {
  }
};

class DPMManager : public Manager {
 protected:
  void UpdateHideApp(shared_ptr<item::AbstractItem> updatedItem) {
    string hide_list;
    int uid = static_pointer_cast<IItemInfoInternal>(updatedItem->GetInfo())->GetUid();
    string map_key = updatedItem->GetId() + updatedItem->GetSenderAppId() + string(to_string(uid));
    list<string> updated_hide_list =
      static_pointer_cast<item::IItemInfoInternal>(updatedItem->GetInfo())->GetHideViewerList();

    auto noti = facade_->hide_map_.find(map_key);
    if (noti == facade_->hide_map_.end())
      return;

    /* hide app list synchronization */
    hide_list = noti->second;
    istringstream stream(hide_list);
    string hide_app_id;

    while (getline(stream, hide_app_id, ';')) {
      list<string>::iterator iter = std::find(updated_hide_list.begin(),
                                updated_hide_list.end(), hide_app_id);
      if (iter == updated_hide_list.end())
        static_pointer_cast<item::IItemInfoInternal>(updatedItem->GetInfo())->AddHideViewer(hide_app_id);
    }
  }

  void OnAdd(const IEventInfo& info,
      list<shared_ptr<item::AbstractItem>> addedItem) override {
    DBG("Add !!!");
    int ret;

    ret = DBManager::InsertNotification(addedItem);
    if (ret == ERROR_NONE) {
      facade_->DelegateReporterEvent(info, addedItem);
      facade_->LaunchDefaultViewer(addedItem, NOTIFICATION_OP_INSERT);
    } else {
        DBG("SendError !!!");
        SendError(info, static_cast<NotificationError>(ret));
    }
  }

  void OnUpdate(const IEventInfo& info,
      shared_ptr<item::AbstractItem> updatedItem) override {
    DBG("Update !!!");
    int ret;

    UpdateHideApp(updatedItem);

    ret = DBManager::UpdateNotification(updatedItem);
    if (ret == ERROR_NONE) {
      facade_->DelegateReporterEvent(info, updatedItem);
      facade_->LaunchDefaultViewer(updatedItem, NOTIFICATION_OP_UPDATE);
    } else {
      SendError(info, static_cast<NotificationError>(ret));
    }
  }

  int DoDelete(const IEventInfo& info,
      shared_ptr<item::AbstractItem> deletedItem) {
    int ret = DBManager::DeleteNotification(deletedItem);
    if (ret == ERROR_NONE) {
      facade_->DelegateReporterEvent(info, deletedItem);
    } else {
      SendError(info, static_cast<NotificationError>(ret));
    }
    return ret;
  }

  void OnDelete(const IEventInfo& info,
      shared_ptr<item::AbstractItem> deletedItem) override {
    DBG("Delete !!!");
    if (info.GetEventType() == IEventInfo::EventType::DeleteAll) {
      list<shared_ptr<item::AbstractItem>> noti_list =
          DBManager::GetNotificationList(info.GetOwner(),
              ((const IEventInfoInternal&)info).GetUid());
      for (auto& i : noti_list) {
        DoDelete(info, i);
      }
      facade_->LaunchDefaultViewer(noti_list, NOTIFICATION_OP_DELETE_ALL);
    } else {
      int ret = DoDelete(info, deletedItem);
      if (ret == ERROR_NONE)
        facade_->LaunchDefaultViewer(deletedItem, NOTIFICATION_OP_DELETE);
    }
  }

  list<shared_ptr<item::AbstractItem>> OnRequestEvent(const IEventInfo& info) override {
    DBG("Get !!! %s", info.GetOwner().c_str());
    if (info.GetItemId().empty()) {
      /* get */
      return DBManager::GetNotificationList(info.GetOwner(),
                         ((const IEventInfoInternal&)info).GetUid());
    } else {
      /*  FindByRootID */
      return DBManager::GetNotificationList(info.GetOwner(), info.GetItemId(),
                         ((const IEventInfoInternal&)info).GetUid());
    }
  }

 public:
  DPMManager(std::unique_ptr<IEventSender> sender,
      std::unique_ptr<IEventListener> listener)
    : Manager(move(sender), move(listener)) {
  }
};

HAPI int notification_ex_service_init(int restart_count) {
  facade_ = new DPMFacade(
    unique_ptr<Reporter>(
      new DPMReporter(
        unique_ptr<DBusSender>(new DBusSender(Manager::GetPath())),
        unique_ptr<DBusEventListener>(new DBusEventListener(Reporter::GetPath())))),
    unique_ptr<Manager>(
      new DPMManager(
        unique_ptr<DBusSender>(new DBusSender(Reporter::GetPath())),
        unique_ptr<DBusEventListener>(new DBusEventListener(Manager::GetPath())))),
        restart_count);

  notification_init_default_viewer();

  return ERROR_NONE;
}

HAPI int notification_ex_service_fini() {
  delete facade_;
  return ERROR_NONE;
}

HAPI GDBusConnection* notification_ex_service_get_gdbus_connection() {
  return DBusConnectionManager::GetInst().GetConnection();
}
