/*
 * Copyright 2016  Samsung Electronics Co., Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <gio/gio.h>
#include <gio/gunixfdlist.h>

#include <dlog.h>
#include <sys/smack.h>
#include <pkgmgr-info.h>
#include <notification.h>
#include <alarm.h>
#include <aul.h>

#include "pkgmgr.h"
#include "service_common.h"
#include "notification_service.h"
#include "debug.h"

#include <notification_noti.h>
#include <notification_internal.h>
#include <notification_ipc.h>
#include <notification_setting_service.h>
#include <notification_db.h>
#include <notification_viewer.h>
#include <notification_shared_file.h>

#define PROVIDER_NOTI_INTERFACE_NAME "org.tizen.data_provider_noti_service"
#define PROVIDER_NOTI_EVENT_INTERFACE_NAME "org.tizen.data_provider_noti_event_service"
#define NOTI_TEMPLATE_LIMIT 10
#define BUF_LEN 256

typedef struct _dnd_alarm_id dnd_alarm_id_s;

static GHashTable *_monitoring_hash;

static GList *_dnd_alarm_id_list;

typedef struct _dnd_alarm_id {
	uid_t uid;
	alarm_id_t dnd_start_id;
	alarm_id_t dnd_end_id;
} dnd_alarm_id_s;

static GList *__event_list;

typedef struct _event_sender_info {
	int priv_id;
	char *busname;
	guint watcher_id;
} event_sender_info_s;

static int _update_noti(GVariant **reply_body, notification_h noti, uid_t uid);
static int _delete_noti(const char *app_id, int priv_id, uid_t uid);

/*!
 * SERVICE HANDLER
 */

/*!
 * NOTIFICATION SERVICE INITIALIZATION
 */
static void __print_noti(notification_h noti)
{
	char *pkgname = NULL;
	char *text = NULL;
	char *content = NULL;
	const char *tag = NULL;

	notification_get_pkgname(noti, &pkgname);
	notification_get_text(noti, NOTIFICATION_TEXT_TYPE_TITLE, &text);
	notification_get_text(noti, NOTIFICATION_TEXT_TYPE_CONTENT, &content);
	notification_get_tag(noti, &tag);

	DBG("Noti-info : pkgname[%s] title[%s] content[%s] tag[%s]",
		pkgname, text, content, tag);
}

static void _on_name_appeared(GDBusConnection *connection,
		const gchar     *name,
		const gchar     *name_owner,
		gpointer         user_data)
{
	monitoring_info_s *info = (monitoring_info_s *)user_data;

	DBG("name[%s], uid[%d]", name, info->uid);
}

static void _on_name_vanished(GDBusConnection *connection,
		const gchar     *name,
		gpointer         user_data)
{
	monitoring_info_s *info = (monitoring_info_s *)user_data;

	if (info) {
		DBG("name[%s], uid[%d]", name, info->uid);
		notification_remove_private_sharing_target_id(name, info->uid);
		g_bus_unwatch_name(info->watcher_id);
		delete_monitoring_list(&_monitoring_hash, info->bus_name, info->uid);

		if (info->bus_name)
			free(info->bus_name);
		free(info);
	}
}

static void _noti_dbus_method_call_handler(GDBusConnection *conn,
		const gchar *sender, const gchar *object_path,
		const gchar *iface_name, const gchar *method_name,
		GVariant *parameters, GDBusMethodInvocation *invocation,
		gpointer user_data)
{
	/* TODO : sender authority(privilege) check */
	INFO("notification method_name[%s] sender[%s]", method_name, sender);

	GVariant *reply_body = NULL;
	int ret = NOTIFICATION_ERROR_INVALID_OPERATION;
	uid_t uid = get_sender_uid(sender);
	pid_t pid = get_sender_pid(sender);

	if (g_strcmp0(method_name, "noti_service_register") == 0) {
		ret = service_register(parameters, &reply_body, sender,
				_on_name_appeared, _on_name_vanished, &_monitoring_hash, uid);
		if (ret == NOTIFICATION_ERROR_NONE)
			notification_add_private_sharing_target_id(pid, sender, uid);
	} else if (g_strcmp0(method_name, "update_noti") == 0) {
		ret = notification_update_noti(parameters, &reply_body, pid, uid);
	} else if (g_strcmp0(method_name, "add_noti") == 0) {
		ret = notification_add_noti(parameters, &reply_body, sender, pid, uid);
	} else if (g_strcmp0(method_name, "refresh_noti") == 0) {
		ret = notification_refresh_noti(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "del_noti_single") == 0) {
		ret = notification_del_noti_single(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "del_noti_multiple") == 0) {
		ret = notification_del_noti_multiple(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "get_noti_count") == 0) {
		ret = notification_get_noti_count(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "get_noti_all_count") == 0) {
		ret = notification_get_noti_all_count(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "update_noti_setting") == 0) {
		ret = notification_update_noti_setting(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "update_noti_sys_setting") == 0) {
		ret = notification_update_noti_sys_setting(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "load_noti_by_tag") == 0) {
		ret = notification_load_noti_by_tag(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "load_noti_by_priv_id") == 0) {
		ret = notification_load_noti_by_priv_id(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "load_noti_grouping_list") == 0) {
		notification_add_private_sharing_target_id(pid, sender, uid);
		notification_load_grouping_list(invocation, parameters, uid);
		return;
	} else if (g_strcmp0(method_name, "load_noti_detail_list") == 0) {
		notification_add_private_sharing_target_id(pid, sender, uid);
		ret = notification_load_detail_list(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "get_setting_array") == 0) {
		ret = notification_get_setting_array(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "get_setting_by_app_id") == 0) {
		ret = notification_get_setting_by_app_id(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "load_system_setting") == 0) {
		ret = notification_load_system_setting(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "save_as_template") == 0) {
		ret = notification_add_noti_template(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "create_from_template") == 0) {
		ret = notification_get_noti_template(parameters, &reply_body, pid, uid);
	} else if (g_strcmp0(method_name, "create_from_package_template") == 0) {
		ret = notification_get_noti_package_template(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "get_noti_block_state") == 0) {
		ret = notification_get_block_state(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "load_dnd_allow_exception") == 0) {
		ret = notification_load_dnd_allow_exception(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "update_dnd_allow_exception") == 0) {
		ret = notification_update_dnd_allow_exception(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "send_noti_event") == 0) {
		ret = notification_send_noti_event(parameters, &reply_body);
	} else if (g_strcmp0(method_name, "send_noti_event_by_priv_id") == 0) {
		ret = notification_send_noti_event_by_priv_id(parameters, &reply_body);
	} else if (g_strcmp0(method_name, "check_event_receiver") == 0) {
		ret = notification_check_event_receiver(parameters, &reply_body);
	} else if (g_strcmp0(method_name, "reset_event_handler") == 0) {
		ret = notification_reset_event_receiver(parameters, &reply_body, sender);
	} else if (g_strcmp0(method_name, "del_noti_by_display_applist") == 0) {
		ret = notification_del_noti_by_display_applist(parameters, &reply_body, uid);
	}

	if (ret == NOTIFICATION_ERROR_NONE) {
		INFO("notification service success, method name[%s]", method_name);
		g_dbus_method_invocation_return_value(
				invocation, reply_body);
	} else {
		ERR("notification service fail, method_name[%s] err[%d]",
			method_name, ret);
		g_dbus_method_invocation_return_error(
				invocation,
				NOTIFICATION_ERROR,
				ret,
				"notification service error");
	}
}

static const GDBusInterfaceVTable _noti_interface_vtable = {
		_noti_dbus_method_call_handler,
		NULL,
		NULL
};

int notification_register_dbus_interface(void)
{
	static gchar introspection_xml[] =
			"  <node>"
			"  <interface name='org.tizen.data_provider_noti_service'>"
			"        <method name='noti_service_register'>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='add_noti'>"
			"          <arg type='v' name='noti' direction='in'/>"
			"          <arg type='i' name='priv_id' direction='out'/>"
			"        </method>"

			"        <method name='update_noti'>"
			"          <arg type='v' name='noti' direction='in'/>"
			"          <arg type='i' name='priv_id' direction='out'/>"
			"        </method>"

			"        <method name='refresh_noti'>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='del_noti_single'>"
			"          <arg type='s' name='app_id' direction='in'/>"
			"          <arg type='i' name='priv_id' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='i' name='priv_id' direction='out'/>"
			"        </method>"

			"        <method name='del_noti_multiple'>"
			"          <arg type='s' name='app_id' direction='in'/>"
			"          <arg type='i' name='priv_id' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='i' name='priv_id' direction='out'/>"
			"        </method>"

			"        <method name='load_noti_by_tag'>"
			"          <arg type='s' name='app_id' direction='in'/>"
			"          <arg type='s' name='tag' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='v' name='noti' direction='out'/>"
			"        </method>"

			"        <method name='load_noti_by_priv_id'>"
			"          <arg type='s' name='app_id' direction='in'/>"
			"          <arg type='i' name='priv_id' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='v' name='noti' direction='out'/>"
			"        </method>"

			"        <method name='load_noti_grouping_list'>"
			"          <arg type='i' name='type' direction='in'/>"
			"          <arg type='i' name='count' direction='in'/>"
			"          <arg type='i' name='count_per_page' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='i' name='list_count' direction='out'/>"
			"        </method>"

			"        <method name='load_noti_detail_list'>"
			"          <arg type='s' name='app_id' direction='in'/>"
			"          <arg type='i' name='group_id' direction='in'/>"
			"          <arg type='i' name='priv_id' direction='in'/>"
			"          <arg type='i' name='count' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='a(v)' name='noti_list' direction='out'/>"
			"        </method>"

			"        <method name='get_noti_count'>"
			"          <arg type='i' name='type' direction='in'/>"
			"          <arg type='s' name='app_id' direction='in'/>"
			"          <arg type='i' name='group_id' direction='in'/>"
			"          <arg type='i' name='priv_id' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='i' name='ret_count' direction='out'/>"
			"        </method>"

			"        <method name='get_noti_all_count'>"
			"          <arg type='i' name='type' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='i' name='ret_count' direction='out'/>"
			"        </method>"

			"        <method name='update_noti_setting'>"
			"          <arg type='s' name='pkgname' direction='in'/>"
			"          <arg type='s' name='app_id' direction='in'/>"
			"          <arg type='i' name='allow_to_notify' direction='in'/>"
			"          <arg type='i' name='do_not_disturb_except' direction='in'/>"
			"          <arg type='i' name='visibility_class' direction='in'/>"
			"          <arg type='i' name='pop_up_notification' direction='in'/>"
			"          <arg type='i' name='lock_screen_content_level' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='update_noti_sys_setting'>"
			"          <arg type='i' name='do_not_disturb' direction='in'/>"
			"          <arg type='i' name='visibility_class' direction='in'/>"
			"          <arg type='i' name='dnd_schedule_enabled' direction='in'/>"
			"          <arg type='i' name='dnd_schedule_day' direction='in'/>"
			"          <arg type='i' name='dnd_start_hour' direction='in'/>"
			"          <arg type='i' name='dnd_start_min' direction='in'/>"
			"          <arg type='i' name='dnd_end_hour' direction='in'/>"
			"          <arg type='i' name='dnd_end_min' direction='in'/>"
			"          <arg type='i' name='lock_screen_level' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='get_setting_array'>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='i' name='setting_cnt' direction='out'/>"
			"          <arg type='a(v)' name='setting_arr' direction='out'/>"
			"        </method>"

			"        <method name='get_setting_by_app_id'>"
			"          <arg type='s' name='app_id' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='v' name='setting' direction='out'/>"
			"        </method>"

			"        <method name='load_system_setting'>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='v' name='setting' direction='out'/>"
			"        </method>"

			"        <method name='save_as_template'>"
			"          <arg type='v' name='noti' direction='in'/>"
			"          <arg type='s' name='name' direction='in'/>"
			"        </method>"

			"        <method name='create_from_template'>"
			"          <arg type='s' name='name' direction='in'/>"
			"          <arg type='v' name='noti' direction='out'/>"
			"        </method>"

			"        <method name='create_from_package_template'>"
			"          <arg type='s' name='app_id' direction='in'/>"
			"          <arg type='s' name='name' direction='in'/>"
			"          <arg type='v' name='noti' direction='out'/>"
			"        </method>"

			"        <method name='get_noti_block_state'>"
			"          <arg type='s' name='app_id' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='v' name='state' direction='out'/>"
			"        </method>"

			"        <method name='update_dnd_allow_exception'>"
			"          <arg type='i' name='type' direction='in'/>"
			"          <arg type='i' name='value' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='load_dnd_allow_exception'>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='i' name='count' direction='out'/>"
			"          <arg type='a(v)' name='dnd_allow_exception' direction='out'/>"
			"        </method>"

			"        <method name='send_noti_event'>"
			"          <arg type='v' name='noti' direction='in'/>"
			"          <arg type='i' name='event_type' direction='in'/>"
			"        </method>"

			"        <method name='send_noti_event_by_priv_id'>"
			"          <arg type='i' name='priv_id' direction='in'/>"
			"          <arg type='i' name='event_type' direction='in'/>"
			"        </method>"

			"        <method name='check_event_receiver'>"
			"          <arg type='i' name='priv_id' direction='in'/>"
			"          <arg type='i' name='available' direction='out'/>"
			"        </method>"

			"        <method name='reset_event_handler'>"
			"          <arg type='i' name='priv_id' direction='in'/>"
			"        </method>"

			"        <method name='del_noti_by_display_applist'>"
			"          <arg type='i' name='display_applist' direction='in'/>"
			"          <arg type='i' name='uid' direction='in'/>"
			"          <arg type='i' name='priv_id' direction='out'/>"
			"        </method>"

			"        <method name='post_toast'>"
			"        </method>"
			"  </interface>"
			"  </node>";

	return service_common_register_dbus_interface(introspection_xml, _noti_interface_vtable);
}

static void __free_event_info(gpointer data)
{
	event_sender_info_s *info = (event_sender_info_s *)data;

	if (info) {
		if (info->busname)
			free(info->busname);
		free(info);
	}
}

static gint __priv_id_compare(gconstpointer a, gconstpointer b)
{
	event_sender_info_s *info = NULL;

	if (!a)
		return -1;

	info = (event_sender_info_s *)a;

	if (info->priv_id == GPOINTER_TO_INT(b))
		return 0;

	return 1;
}

static int __check_limit(notification_h noti, uid_t uid)
{
	int ret;
	int priv_id = NOTIFICATION_PRIV_ID_NONE;
	char *app_id = NULL;
	GList *list = NULL;
	GList *iter_list = NULL;

	ret = notification_noti_check_limit(noti, uid, &list);
	if (ret == NOTIFICATION_ERROR_NONE && list != NULL) {
		ret = notification_get_pkgname(noti, &app_id);
		if (ret == NOTIFICATION_ERROR_NONE && app_id != NULL) {
			iter_list = list;
			while (iter_list) {
				priv_id = GPOINTER_TO_INT(iter_list->data);

				ret = _delete_noti((const char *)app_id, priv_id, uid);
				if (ret != NOTIFICATION_ERROR_NONE)
					ERR("Failed to delete noti[%d], err[%d]",
						priv_id, ret);

				iter_list = g_list_next(iter_list);
			}
		}
	}

	if (list)
		g_list_free(list);

	return ret;
}

/* add noti */
static int _add_noti(GVariant **reply_body, notification_h noti, uid_t uid)
{
	int ret;
	int priv_id = NOTIFICATION_PRIV_ID_NONE;
	GVariant *body = NULL;

	__print_noti(noti);
	ret = notification_noti_insert(noti);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to update a notification [%d]", ret);
		return ret;
	}

	ret = notification_get_id(noti, NULL, &priv_id);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to gets priv_id [%d]", ret);
		return ret;
	}

	notification_set_private_sharing(noti, uid);

	body = notification_ipc_make_gvariant_from_noti(noti, true);
	if (body == NULL) {
		ERR("Failed to make gvariant to noti");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	ret = send_notify(body, "add_noti_notify", &_monitoring_hash,
			PROVIDER_NOTI_INTERFACE_NAME, uid);
	g_variant_unref(body);

	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to send notify [%d]", ret);
		return ret;
	}

	ret = notification_launch_default_viewer(priv_id,
			NOTIFICATION_OP_INSERT, uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to launch default viewer [%d]", ret);
		ret = NOTIFICATION_ERROR_NONE;
	}


	*reply_body = g_variant_new("(i)", priv_id);
	if (*reply_body == NULL) {
		ERR("Failed to make reply");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	ret = __check_limit(noti, uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to check limit [%d]", ret);
		ret = NOTIFICATION_ERROR_NONE;
	}

	DBG("Success to add notification");
	return ret;
}

static int _validate_and_set_noti_with_uid(uid_t uid, notification_h noti, uid_t *noti_uid)
{
	int ret = NOTIFICATION_ERROR_NONE;

	ret = notification_get_uid(noti, noti_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to get uid [%d]", ret);
		return ret;
	}

	if (uid > NORMAL_UID_BASE && uid != *noti_uid) {
		ERR("Invalid seder uid[%d] noti_uid[%d]", uid, *noti_uid);
		return NOTIFICATION_ERROR_INVALID_PARAMETER;
	} else if (uid <= NORMAL_UID_BASE) {
		if (*noti_uid <= NORMAL_UID_BASE) {
			if (*noti_uid != uid)
				return NOTIFICATION_ERROR_INVALID_PARAMETER;

			DBG("system defualt noti post change noti_uid to default");
			/*
			IF system (uid <= NORMAL_UID_BASE) try to send noti without specipic destination uid (using notification_pot API)
			Send it to default user (TZ_SYS_DEFAULT_USER)
			*/
			*noti_uid = tzplatform_getuid(TZ_SYS_DEFAULT_USER);
			ret = notification_set_uid(noti, *noti_uid);
			if (ret != NOTIFICATION_ERROR_NONE) {
				ERR("Failed to set uid [%d]", ret);
				return ret;
			}
			DBG("changed noti_uid[%d]", *noti_uid);
		}
	}
	return ret;
}

static int _validate_and_set_param_uid_with_uid(uid_t uid, uid_t *param_uid)
{
	if (uid > NORMAL_UID_BASE && uid != *param_uid) {
		ERR("invalid seder uid[%d], param_uid[%d]", uid, *param_uid);
		return NOTIFICATION_ERROR_INVALID_PARAMETER;
	} else if (uid <= NORMAL_UID_BASE) {
		if (*param_uid <= NORMAL_UID_BASE) {
			if (*param_uid != uid)
				return NOTIFICATION_ERROR_INVALID_PARAMETER;
			/*
			IF system (uid <= NORMAL_UID_BASE) try to send noti without specipic destination uid (using notification_pot API)
			Send it to default user (TZ_SYS_DEFAULT_USER)
			*/
			*param_uid = tzplatform_getuid(TZ_SYS_DEFAULT_USER);
		}
	}
	return NOTIFICATION_ERROR_NONE;
}

static void __sender_name_appeared_cb(GDBusConnection *connection,
				      const gchar *name,
				      const gchar *name_owner,
				      gpointer user_data)
{
	DBG("name[%s], name_owner[%s]", name, name_owner);
}

static void __sender_name_vanished_cb(GDBusConnection *connection,
				      const gchar *name,
				      gpointer user_data)
{
	GList *delete_list = NULL;
	event_sender_info_s *info;

	if (!user_data)
		return;

	info = (event_sender_info_s *)user_data;

	DBG("name[%s], priv_id[%d], watcher_id[%d]",
		 name, info->priv_id, info->watcher_id);

	delete_list = g_list_find_custom(__event_list, GINT_TO_POINTER(info->priv_id),
					 (GCompareFunc)__priv_id_compare);
	if (delete_list) {
		__event_list = g_list_remove(__event_list, info);
		g_bus_unwatch_name(info->watcher_id);
		__free_event_info(info);
	}
}

static guint __insert_sender_watcher_id(event_sender_info_s *sender_info)
{
	guint watcher_id = 0;

	watcher_id = g_bus_watch_name(G_BUS_TYPE_SYSTEM,
				      sender_info->busname,
				      G_BUS_NAME_WATCHER_FLAGS_NONE,
				      __sender_name_appeared_cb,
				      __sender_name_vanished_cb,
				      sender_info,
				      NULL);
	if (!watcher_id) {
		ERR("Failed to watch sender name");
		return 0;
	}

	DBG("Watch on busname[%s] watcher_id[%d]", sender_info->busname, watcher_id);
	return watcher_id;
}

static void __add_sender_info(int priv_id, const char *busname)
{
	GList *find_list = NULL;
	event_sender_info_s *sender_info;
	event_sender_info_s *tmp_info = NULL;

	if (busname == NULL)
		return;

	sender_info = (event_sender_info_s *)malloc(sizeof(event_sender_info_s));
	if (sender_info == NULL) {
		ERR("malloc failed");
		return;
	}

	sender_info->priv_id = priv_id;
	sender_info->busname = strdup(busname);

	if (__event_list == NULL) {
		sender_info->watcher_id = __insert_sender_watcher_id(sender_info);
		__event_list = g_list_append(__event_list, sender_info);
	} else {
		__event_list = g_list_first(__event_list);
		find_list = g_list_find_custom(__event_list, GINT_TO_POINTER(priv_id),
					       (GCompareFunc)__priv_id_compare);
		if (find_list == NULL) {
			sender_info->watcher_id = __insert_sender_watcher_id(sender_info);
			__event_list = g_list_append(__event_list, sender_info);
		} else {
			tmp_info = (event_sender_info_s *)g_list_nth_data(find_list, 0);
			if (tmp_info->busname) {
				if (strcmp(tmp_info->busname, busname)) {
					free(tmp_info->busname);
					tmp_info->busname = strdup(busname);
					g_bus_unwatch_name(tmp_info->watcher_id);
					tmp_info->watcher_id = __insert_sender_watcher_id(sender_info);
				}
			} else {
				tmp_info->busname = strdup(busname);
				if (!tmp_info->watcher_id)
					g_bus_unwatch_name(tmp_info->watcher_id);
				tmp_info->watcher_id = __insert_sender_watcher_id(tmp_info);
			}
			__free_event_info(sender_info);
		}
	}
}

int notification_add_noti(GVariant *parameters, GVariant **reply_body,
		const char *sender, pid_t pid, uid_t uid)
{
	int ret;
	int priv_id;
	bool event_flag;
	notification_h noti;
	GVariant *body = NULL;
	uid_t noti_uid = 0;

	noti = notification_create(NOTIFICATION_TYPE_NOTI);
	if (noti != NULL) {
		g_variant_get(parameters, "(v)", &body);
		ret = notification_ipc_make_noti_from_gvariant(noti, body);
		g_variant_unref(body);

		if (uid >= NORMAL_UID_BASE)
			notification_set_indirect_request(noti, pid, uid);

		ret = _validate_and_set_noti_with_uid(uid, noti, &noti_uid);
		if (ret != NOTIFICATION_ERROR_NONE)
			goto out;

		ret = notification_noti_check_tag(noti);
		if (ret == NOTIFICATION_ERROR_NOT_EXIST_ID)
			ret = _add_noti(reply_body, noti, noti_uid);
		else if (ret == NOTIFICATION_ERROR_ALREADY_EXIST_ID)
			ret = _update_noti(reply_body, noti, noti_uid);

		if (ret != NOTIFICATION_ERROR_NONE)
			goto out;

		ret = notification_get_event_flag(noti, &event_flag);
		if (ret != NOTIFICATION_ERROR_NONE)
			goto out;

		if (event_flag == true) {
			ret = notification_get_id(noti, NULL, &priv_id);
			if (ret != NOTIFICATION_ERROR_NONE)
				goto out;

			__add_sender_info(priv_id, sender);
		}
	} else {
		ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

out:
	notification_free(noti);

	return ret;
}

/* update noti */
static int _update_noti(GVariant **reply_body, notification_h noti, uid_t uid)
{
	int ret;
	GVariant *body = NULL;
	int priv_id = NOTIFICATION_PRIV_ID_NONE;
	notification_h source_noti;

	__print_noti(noti);

	ret = notification_get_id(noti, NULL, &priv_id);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to get priv_id [%d]", ret);
		return ret;
	}

	if (notification_validate_private_sharing(noti) == false) {
		source_noti = notification_create(NOTIFICATION_TYPE_NOTI);
		if (source_noti == NULL)
			return get_last_result();
		notification_noti_get_by_priv_id(source_noti, priv_id);

		notification_calibrate_private_sharing(noti, source_noti);
		notification_free(source_noti);
	}

	ret = notification_noti_update(noti);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to update a notification [%d]", ret);
		return ret;
	}

	notification_set_private_sharing(noti, uid);

	body = notification_ipc_make_gvariant_from_noti(noti, true);
	if (body == NULL) {
		ERR("Failed to make gvariant to noti");
		return NOTIFICATION_ERROR_IO_ERROR;
	}

	ret = send_notify(body, "update_noti_notify", &_monitoring_hash, PROVIDER_NOTI_INTERFACE_NAME, uid);
	g_variant_unref(body);

	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to send notify [%d]", ret);
		return ret;
	}

	ret = notification_launch_default_viewer(priv_id,
			NOTIFICATION_OP_UPDATE, uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to launch default viewer [%d]", ret);
		ret = NOTIFICATION_ERROR_NONE;
	}

	*reply_body = g_variant_new("(i)", priv_id);
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	DBG("Success to update noti");
	return ret;
}

int notification_update_noti(GVariant *parameters, GVariant **reply_body,
		pid_t pid, uid_t uid)
{
	notification_h noti;
	int ret;
	GVariant *body = NULL;
	uid_t noti_uid;

	noti = notification_create(NOTIFICATION_TYPE_NOTI);
	if (noti != NULL) {
		g_variant_get(parameters, "(v)", &body);
		ret = notification_ipc_make_noti_from_gvariant(noti, body);
		g_variant_unref(body);

		if (uid >= NORMAL_UID_BASE)
			notification_set_indirect_request(noti, pid, uid);

		ret = _validate_and_set_noti_with_uid(uid, noti, &noti_uid);
		if (ret != NOTIFICATION_ERROR_NONE)
			goto out;

		ret = _update_noti(reply_body, noti, noti_uid);

	} else {
		ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

out:
	notification_free(noti);
	return ret;
}

/* load_noti_by_tag */
int notification_load_noti_by_tag(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	char *tag = NULL;
	char *app_id = NULL;
	notification_h noti;
	uid_t param_uid;

	noti = notification_create(NOTIFICATION_TYPE_NOTI);
	if (noti != NULL) {
		g_variant_get(parameters, "(&s&si)", &app_id, &tag, &param_uid);
		ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
		if (ret != NOTIFICATION_ERROR_NONE)
			goto out;

		DBG("_load_noti_by_tag app_id : %s, tag : %s ", app_id, tag);
		ret = notification_noti_get_by_tag(noti, app_id, tag, param_uid);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("No matched tag [%s]", tag);
			goto out;
		}

		__print_noti(noti);
		*reply_body = notification_ipc_make_gvariant_from_noti(noti, true);
		if (*reply_body == NULL) {
			ERR("Failed to make reply_body");
			ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
		}
	} else {
		ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

out:
	if (noti)
		notification_free(noti);

	return ret;
}

/* load_noti_by_priv_id */
int notification_load_noti_by_priv_id(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	int priv_id = NOTIFICATION_PRIV_ID_NONE;
	char *app_id = NULL;
	notification_h noti;
	uid_t param_uid;

	noti = notification_create(NOTIFICATION_TYPE_NOTI);
	if (noti != NULL) {
		g_variant_get(parameters, "(&sii)", &app_id, &priv_id, &param_uid);
		ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
		if (ret != NOTIFICATION_ERROR_NONE)
			goto out;

		DBG("load_noti_by_priv_id app_id : %s, priv_id : %d ", app_id, priv_id);
		ret = notification_noti_get_by_priv_id(noti, priv_id);
		if (ret != NOTIFICATION_ERROR_NONE)
			goto out;

		*reply_body = notification_ipc_make_gvariant_from_noti(noti, true);
		if (*reply_body == NULL) {
			ERR("Failed to make reply_body");
			ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
		}
	} else {
		ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

out:
	if (noti)
		notification_free(noti);

	return ret;
}

static int __get_socket_from_invocation(GDBusMethodInvocation *invocation, int *fd)
{
	int ret_fd = 0;
	GError *g_err = NULL;
	GDBusMessage *msg = NULL;
	GUnixFDList *fd_list = NULL;

	msg = g_dbus_method_invocation_get_message(invocation);
	fd_list = g_dbus_message_get_unix_fd_list(msg);
	if (fd_list == NULL) {
		ERR("fd list is null");
		return NOTIFICATION_ERROR_IO_ERROR;
	}

	ret_fd = g_unix_fd_list_get(fd_list, 0, &g_err);
	if (ret_fd == -1 || g_err) {
		ERR("g_unix_fd_list_get [%s]", g_err->message);
		g_error_free(g_err);
		return NOTIFICATION_ERROR_IO_ERROR;
	}

	*fd = ret_fd;

	INFO("sender fd : %d", ret_fd);
	return NOTIFICATION_ERROR_NONE;
}

static int __send_list_to_socket(int fd, notification_list_h get_list, uid_t uid)
{
	int ret;
	notification_h noti;
	notification_list_h list_iter;
	GVariant *body = NULL;

	list_iter = notification_list_get_head(get_list);
	do {
		noti = notification_list_get_data(list_iter);
		notification_set_private_sharing(noti, uid);
		body = notification_ipc_make_gvariant_from_noti(noti, true);
		if (body == NULL) {
			ERR("Failed to make gvariant to noti");
			ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
			break;
		}
		ret = notification_ipc_socket_write_string(fd,
				(char *)g_variant_get_data(body),
				g_variant_get_size(body));
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("write string to socket fail");
			g_variant_unref(body);
			return ret;
		}
		list_iter = notification_list_get_next(list_iter);
		g_variant_unref(body);
	} while (list_iter != NULL);

	return ret;
}

/* load_noti_grouping_list */
void notification_load_grouping_list(GDBusMethodInvocation *invocation, GVariant *parameters, uid_t uid)
{
	int ret;
	notification_type_e type;
	notification_list_h get_list = NULL;
	GVariant *reply_body = NULL;
	int fd = 0;
	int count = 0;
	int count_per_page;
	int list_count = 0;
	uid_t param_uid;

	g_variant_get(parameters, "(iiii)", &type, &count, &count_per_page, &param_uid);
	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		goto err;

	ret = __get_socket_from_invocation(invocation, &fd);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("get_fd_from_invocation [%d]", ret);
		goto err;
	}

	ret = notification_noti_get_grouping_list(type, count, count_per_page,
			&get_list, &list_count, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("notification_noti_get_grouping_list [%d]", ret);
		goto err;
	}

	reply_body = g_variant_new("(i)", list_count);
	if (reply_body == NULL) {
		ERR("Failed to make reply_body");
		ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
		goto err;
	}

	g_dbus_method_invocation_return_value(invocation, reply_body);
	if (get_list) {
		ret = __send_list_to_socket(fd, get_list, param_uid);
		if (ret != NOTIFICATION_ERROR_NONE)
			ERR("__send_list_to_socket [%d]", ret);
		notification_free_list(get_list);
	}

	close(fd);
	return;

err:
	if (fd)
		close(fd);
	if (get_list)
		notification_free_list(get_list);

	ERR("notification service fail, method name[load_noti_grouping_list] [%d]", ret);
	g_dbus_method_invocation_return_error(invocation, NOTIFICATION_ERROR,
				ret, "notification_service_error");
}

/* get_setting_array */
int notification_get_setting_array(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	GVariant *body;
	GVariantBuilder *builder;
	int count = 0;
	int i;
	notification_setting_h setting_array = NULL;
	notification_setting_h temp;
	uid_t param_uid;

	g_variant_get(parameters, "(i)", &param_uid);
	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = noti_setting_get_setting_array(&setting_array, &count, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	builder = g_variant_builder_new(G_VARIANT_TYPE("a(v)"));

	if (setting_array) {
		for (i = 0; i < count; i++) {
			temp = setting_array + i;
			body = notification_ipc_make_gvariant_from_setting(temp);
			g_variant_builder_add(builder, "(v)", body);

			if (temp->package_name)
				free(temp->package_name);
			if (temp->app_id)
				free(temp->app_id);
		}
		free(setting_array);
	}
	*reply_body = g_variant_new("(ia(v))", count, builder);
	g_variant_builder_unref(builder);

	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	INFO("count[%d]", count);
	return ret;
}

static pkgmgrinfo_pkginfo_h __create_pkginfo_by_app_id(const char *app_id, uid_t uid)
{
	int ret_pkgmgr;
	char *pkgname = NULL;
	pkgmgrinfo_appinfo_h app_handle = NULL;
	pkgmgrinfo_pkginfo_h pkg_handle = NULL;

	if (app_id == NULL)
		return NULL;

	ret_pkgmgr = pkgmgrinfo_appinfo_get_usr_appinfo(app_id, uid, &app_handle);
	if (ret_pkgmgr != PMINFO_R_OK) {
		ERR("Failed to pkgmgrinfo_appinfo_get_appinfo [%s][%d]",
			app_id, ret_pkgmgr);
		goto out;
	}

	ret_pkgmgr = pkgmgrinfo_appinfo_get_pkgname(app_handle, &pkgname);
	if (ret_pkgmgr != PMINFO_R_OK) {
		ERR("Failed to pkgmgrinfo_appinfo_get_pkgname [%s][%d]",
			app_id, ret_pkgmgr);
		goto out;
	}

	ret_pkgmgr = pkgmgrinfo_pkginfo_get_usr_pkginfo(pkgname, uid, &pkg_handle);
	if (ret_pkgmgr != PMINFO_R_OK) {
		ERR("Failed to pkgmgrinfo_appinfo_get_pkginfo [%s][%d]",
			app_id, ret_pkgmgr);
		goto out;
	}

out:
	if (app_handle)
		pkgmgrinfo_appinfo_destroy_appinfo(app_handle);

	return pkg_handle;
}

#define NOTIFICATION_PRIVILEGE "http://tizen.org/privilege/notification"
static int __check_privilege_cb(const char *privilege_name, void *is_existed)
{
	if (!g_strcmp0(privilege_name, NOTIFICATION_PRIVILEGE)) {
		*(int *)is_existed = 1;
		return -1;
	}
	return 0;
}

static int __has_notification_privilege(pkgmgrinfo_pkginfo_h handle)
{
	int ret_pkgmgr;
	int is_existed = 0;

	ret_pkgmgr = pkgmgrinfo_pkginfo_foreach_privilege(handle, __check_privilege_cb, &is_existed);
	if (ret_pkgmgr != PMINFO_R_OK) {
		ERR("Failed to pkgmgrinfo_foreach_privilege [%d]",
			ret_pkgmgr);
		return NOTIFICATION_ERROR_IO_ERROR;
	}

	if (is_existed == 0)
		return NOTIFICATION_ERROR_NOT_EXIST_ID;

	return NOTIFICATION_ERROR_NONE;
}

static int __init_setting_handle_by_app_id(const char *app_id,
				notification_setting_h *setting, uid_t uid)
{
	int ret = NOTIFICATION_ERROR_NONE;
	int ret_pkgmgr;
	char *pkgname = NULL;
	notification_setting_h ret_setting = NULL;
	pkgmgrinfo_pkginfo_h handle = NULL;

	if (app_id == NULL || setting == NULL)
		return NOTIFICATION_ERROR_INVALID_PARAMETER;

	handle = __create_pkginfo_by_app_id(app_id, uid);
	if (handle == NULL) {
		ERR("Failed to create pkginfo by app_id [%s]", app_id);
		ret = NOTIFICATION_ERROR_IO_ERROR;
		goto out;
	}

	ret = __has_notification_privilege(handle);
	if (ret != NOTIFICATION_ERROR_NONE) {
		if (ret == NOTIFICATION_ERROR_NOT_EXIST_ID)
			DBG("No notification privilege [%s]", app_id);
		else
			ERR("Failed to check privilege [%d]", ret);
		goto out;
	}

	ret_pkgmgr = pkgmgrinfo_pkginfo_get_pkgname(handle, &pkgname);
	if (ret_pkgmgr != PMINFO_R_OK) {
		ERR("Failed to get pkgname from pkginfo [%d]", ret_pkgmgr);
		ret = NOTIFICATION_ERROR_IO_ERROR;
		goto out;
	}

	ret_setting = (struct notification_setting *)malloc(sizeof(struct notification_setting));
	if (ret_setting == NULL) {
		ERR("Failed to alloc memory");
		ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
		goto out;
	}

	ret_setting->package_name = strdup(pkgname);
	ret_setting->app_id = strdup(app_id);
	ret_setting->allow_to_notify = 1;
	ret_setting->do_not_disturb_except = 0;
	ret_setting->visibility_class = 0;
	ret_setting->pop_up_notification = 1;
	ret_setting->lock_screen_content_level = 0;
	ret_setting->app_disabled = 0;

	*setting = ret_setting;

out:
	if (handle)
		pkgmgrinfo_pkginfo_destroy_pkginfo(handle);

	return ret;
}

/* get_setting_by_app_id */
int notification_get_setting_by_app_id(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	GVariant *body;
	char *app_id = NULL;
	notification_setting_h setting = NULL;
	uid_t param_uid;

	g_variant_get(parameters, "(&si)", &app_id, &param_uid);
	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = noti_setting_service_get_setting_by_app_id(app_id, &setting, param_uid);
	if (ret == NOTIFICATION_ERROR_NOT_EXIST_ID) {
		ret = __init_setting_handle_by_app_id(app_id, &setting, param_uid);
		if (ret != NOTIFICATION_ERROR_NONE)
			return ret;
	} else if (ret != NOTIFICATION_ERROR_NONE) {
		return ret;
	}

	body = notification_ipc_make_gvariant_from_setting(setting);
	notification_setting_free_notification(setting);
	if (body == NULL) {
		ERR("Failed to make gvariant");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	*reply_body = g_variant_new("(v)", body);
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	INFO("app_id[%s] uid[%d]", app_id, param_uid);
	return ret;
}

/* load_system_setting */
int notification_load_system_setting(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	GVariant *body;
	notification_system_setting_h setting = NULL;
	uid_t param_uid;

	g_variant_get(parameters, "(i)", &param_uid);
	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = noti_system_setting_load_system_setting(&setting, param_uid);
	if (ret == NOTIFICATION_ERROR_NONE) {
		body = notification_ipc_make_gvariant_from_system_setting(setting);
		notification_system_setting_free_system_setting(setting);

		if (body == NULL) {
			ERR("Failed to make gvariant");
			return NOTIFICATION_ERROR_OUT_OF_MEMORY;
		}
	} else {
		return ret;
	}

	*reply_body = g_variant_new("(v)", body);
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	return ret;
}

/* load_noti_detail_list */
int notification_load_detail_list(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	notification_h noti;
	GVariant *body = NULL;
	notification_list_h get_list = NULL;
	notification_list_h list_iter;
	GVariantBuilder *builder;
	char *app_id = NULL;
	int group_id = NOTIFICATION_GROUP_ID_NONE;
	int priv_id = NOTIFICATION_PRIV_ID_NONE;
	int count = 0;
	uid_t param_uid;

	g_variant_get(parameters, "(&siiii)", &app_id, &group_id, &priv_id, &count, &param_uid);
	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = notification_noti_get_detail_list(app_id, group_id, priv_id,
			count, &get_list, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	builder = g_variant_builder_new(G_VARIANT_TYPE("a(v)"));

	if (get_list) {
		list_iter = notification_list_get_head(get_list);
		do {
			noti = notification_list_get_data(list_iter);
			notification_set_private_sharing(noti, param_uid);
			body = notification_ipc_make_gvariant_from_noti(noti, true);
			if (body) {
				g_variant_builder_add(builder, "(v)", body);
				list_iter = notification_list_get_next(list_iter);
			}
		} while (list_iter != NULL);
		notification_free_list(get_list);
	}

	*reply_body = g_variant_new("(a(v))", builder);
	g_variant_builder_unref(builder);

	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	DBG("app_id[%s], group_id[%d], priv_id[%d], count[%d]",
			app_id, group_id, priv_id, count);
	return ret;
}

/* refresh_noti */
int notification_refresh_noti(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	uid_t param_uid;

	g_variant_get(parameters, "(i)", &param_uid);
	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = send_notify(parameters, "refresh_noti_notify", &_monitoring_hash, PROVIDER_NOTI_INTERFACE_NAME, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to send notify [%d]", ret);
		return ret;
	}

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	return ret;
}

static int __delete_sender_info(int priv_id)
{
	int ret = NOTIFICATION_ERROR_NONE;
	event_sender_info_s *info;
	GList *delete_list = NULL;
	GVariant *body = NULL;

	if (__event_list != NULL) {
		__event_list = g_list_first(__event_list);
		delete_list = g_list_find_custom(__event_list, GINT_TO_POINTER(priv_id),
						 (GCompareFunc)__priv_id_compare);
		if (delete_list != NULL) {
			info = (event_sender_info_s *)g_list_nth_data(delete_list, 0);
			body = g_variant_new("(i)", priv_id);
			ret = send_event_notify_by_busname(body, "delete_noti", info->busname, PROVIDER_NOTI_EVENT_INTERFACE_NAME);
			g_variant_unref(body);
			if (ret != NOTIFICATION_ERROR_NONE) {
				ERR("Failed to send event notify [%d]", ret);
				return ret;
			}
			__event_list = g_list_remove(g_list_first(__event_list), info);
			g_bus_unwatch_name(info->watcher_id);
			__free_event_info(info);
			if (g_list_length(__event_list) == 0)
				__event_list = NULL;
		}
	}

	return ret;
}

static int _delete_noti(const char *app_id, int priv_id, uid_t uid)
{
	int ret;
	int num_changes = 0;
	GVariant *body = NULL;

	if (app_id == NULL)
		return NOTIFICATION_ERROR_INVALID_PARAMETER;

	ret = notification_noti_delete_by_priv_id_get_changes(app_id, priv_id,
				&num_changes, uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to delete noti, err[%d] num_changes[%d]",
			ret, num_changes);
		return ret;
	}

	if (num_changes > 0) {
		body = g_variant_new("(iii)", 1, priv_id, uid);
		if (body == NULL) {
			ERR("Failed to make gvariant to noti");
			return NOTIFICATION_ERROR_OUT_OF_MEMORY;
		}
		ret = send_notify(body, "delete_single_notify",
			&_monitoring_hash, PROVIDER_NOTI_INTERFACE_NAME, uid);
		g_variant_unref(body);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to send notify [%d]", ret);
			return ret;
		}

		notification_remove_private_sharing(app_id, priv_id, uid);

		ret = __delete_sender_info(priv_id);
		if (ret != NOTIFICATION_ERROR_NONE)
			return ret;

		ret = notification_launch_default_viewer(NOTIFICATION_PRIV_ID_NONE,
				NOTIFICATION_OP_DELETE,
				uid);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to launch default viewer [%d]", ret);
			ret = NOTIFICATION_ERROR_NONE;
		}
	}

	DBG("app_id[%s] priv_id[%d] num_changes[%d]", app_id, priv_id, num_changes);

	return ret;
}

/* del_noti_single */
int notification_del_noti_single(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret = NOTIFICATION_ERROR_NONE;
	int priv_id = NOTIFICATION_PRIV_ID_NONE;
	char *app_id = NULL;
	uid_t param_uid;

	g_variant_get(parameters, "(&sii)", &app_id, &priv_id, &param_uid);
	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = _delete_noti(app_id, priv_id, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to delete noti, err[%d]", ret);
		return ret;
	}

	*reply_body = g_variant_new("(i)", priv_id);
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	return ret;
}

/* del_noti_multiple */
int notification_del_noti_multiple(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	char *app_id = NULL;
	notification_type_e type = NOTIFICATION_TYPE_NONE;
	int num_deleted = 0;
	int *list_deleted = NULL;
	GVariant *deleted_noti_list;
	GVariantBuilder *builder;
	int i;
	uid_t param_uid;

	g_variant_get(parameters, "(&sii)", &app_id, &type, &param_uid);
	DBG("app_id [%s] type [%d]", app_id, type);
	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = notification_noti_delete_all(type, app_id, &num_deleted, &list_deleted, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to delete notifications [%d]", ret);
		if (list_deleted != NULL)
			free(list_deleted);
		return ret;
	}

	if (num_deleted > 0) {
		builder = g_variant_builder_new(G_VARIANT_TYPE("a(i)"));

		for (i = 0; i < num_deleted; i++)
			g_variant_builder_add(builder, "(i)", *(list_deleted + i));

		deleted_noti_list = g_variant_new("(a(i)ii)", builder, num_deleted, param_uid);
		ret = send_notify(deleted_noti_list, "delete_multiple_notify", &_monitoring_hash, PROVIDER_NOTI_INTERFACE_NAME, param_uid);
		g_variant_builder_unref(builder);
		g_variant_unref(deleted_noti_list);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to send notify [%d]", ret);
			return ret;
		}

		for (i = 0; i < num_deleted; i++) {
			notification_remove_private_sharing(app_id, *(list_deleted + i), param_uid);
			ret = __delete_sender_info(*(list_deleted + i));
			if (ret != NOTIFICATION_ERROR_NONE)
				return ret;
		}

		free(list_deleted);

		ret = notification_launch_default_viewer(NOTIFICATION_PRIV_ID_NONE,
			NOTIFICATION_OP_DELETE_ALL,
			param_uid);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to launch default viewer [%d]", ret);
			ret = NOTIFICATION_ERROR_NONE;
		}
	}

	*reply_body = g_variant_new("(i)", num_deleted);
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	INFO("type[%d] app_id[%s] num_deleted[%d]", type, app_id, num_deleted);
	return ret;
}

static void __free_deleted_list_info(notification_deleted_list_info_s *info,
			int count)
{
	int i;

	for (i = 0; i < count; i++) {
		if ((info + i)->app_id != NULL)
			free((info + i)->app_id);
	}

	free(info);
}

int notification_del_noti_by_display_applist(GVariant *parameters,
			GVariant **reply_body, uid_t uid)
{
	int ret;
	int display_applist = 0;
	int deleted_num = 0;
	int i;
	GVariant *deleted_noti_list;
	GVariantBuilder *builder;
	uid_t param_uid;
	notification_deleted_list_info_s *deleted_list = NULL;

	g_variant_get(parameters, "(ii)", &display_applist, &param_uid);

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = notification_noti_delete_by_display_applist(display_applist,
				&deleted_num, &deleted_list, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to delete notifications [%d]", ret);
		if (deleted_list != NULL)
			__free_deleted_list_info(deleted_list, deleted_num);
		return ret;
	}

	if (deleted_num > 0) {
		builder = g_variant_builder_new(G_VARIANT_TYPE("a(i)"));

		for (i = 0; i < deleted_num; i++) {
			g_variant_builder_add(builder, "(i)",
					(deleted_list +i)->priv_id);
			notification_remove_private_sharing(
					(deleted_list +i)->app_id,
					(deleted_list +i)->priv_id,
					param_uid);
			__delete_sender_info((deleted_list +i)->priv_id);
		}
		__free_deleted_list_info(deleted_list, deleted_num);

		deleted_noti_list = g_variant_new("(a(i)ii)", builder,
						deleted_num, param_uid);
		g_variant_builder_unref(builder);

		ret = send_notify(deleted_noti_list,
			"delete_by_display_applist_notify", &_monitoring_hash,
			PROVIDER_NOTI_INTERFACE_NAME, param_uid);
		g_variant_unref(deleted_noti_list);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to send notify [%d]", ret);
			return ret;
		}

		ret = notification_launch_default_viewer(NOTIFICATION_PRIV_ID_NONE,
			NOTIFICATION_OP_DELETE_ALL,
			param_uid);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to launch default viewer [%d]", ret);
			ret = NOTIFICATION_ERROR_NONE;
		}
	}

	*reply_body = g_variant_new("(i)", deleted_num);
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	INFO("deleted num[%d]", deleted_num);
	return ret;
}

/* get_noti_count */
int notification_get_noti_count(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	notification_type_e type = NOTIFICATION_TYPE_NONE;
	char *app_id = NULL;
	int group_id = NOTIFICATION_GROUP_ID_NONE;
	int priv_id = NOTIFICATION_PRIV_ID_NONE;
	int noti_count = 0;
	uid_t param_uid;

	g_variant_get(parameters, "(i&siii)", &type, &app_id, &group_id, &priv_id, &param_uid);
	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to validate uid [%d]", ret);
		return NOTIFICATION_ERROR_IO_ERROR;
	}

	ret = notification_noti_get_count(type, app_id, group_id, priv_id,
			&noti_count, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to get count [%d]", ret);
		return ret;
	}

	*reply_body = g_variant_new("(i)", noti_count);
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	INFO("type[%d] app_id[%s] group_id[%d] priv_id[%d] noti_count[%d]",
		type, app_id, group_id, priv_id, noti_count);
	return ret;
}

/* get_noti_count */
int notification_get_noti_all_count(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	int noti_count = 0;
	int type;
	uid_t param_uid;

	g_variant_get(parameters, "(ii)", &type, &param_uid);
	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to validate uid [%d]", ret);
		return NOTIFICATION_ERROR_IO_ERROR;
	}

	ret = notification_noti_get_all_count(type, &noti_count, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to get count [%d]", ret);
		return ret;
	}

	*reply_body = g_variant_new("(i)", noti_count);
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	INFO("The numbers of all notification count is [%d]", noti_count);
	return ret;
}

/* update_noti_setting */
int notification_update_noti_setting(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	char *pkgname = NULL;
	char *app_id = NULL;
	int allow_to_notify = 0;
	int do_not_disturb_except = 0;
	int visivility_class = 0;
	int pop_up_notification = 0;
	int lock_screen_content_level = 0;
	uid_t param_uid;

	g_variant_get(parameters, "(&s&siiiiii)",
			&pkgname,
			&app_id,
			&allow_to_notify,
			&do_not_disturb_except,
			&visivility_class,
			&pop_up_notification,
			&lock_screen_content_level,
			&param_uid);

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = notification_setting_db_update(pkgname, app_id, allow_to_notify, do_not_disturb_except, visivility_class,
					     pop_up_notification, lock_screen_content_level, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to setting db update : %d", ret);
		return ret;
	}

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	INFO("package_name[%s] app_id[%s] allow_to_notify[%d] do_not_disturb_except[%d] visivility_class[%d] pop_up_notification[%d] lock_screen_content_level[%d]",
			pkgname, app_id, allow_to_notify, do_not_disturb_except, visivility_class, pop_up_notification, lock_screen_content_level);
	return ret;
}

static int _dnd_schedule_alarm_cb(alarm_id_t alarm_id, void *data)
{
	int ret;
	int do_not_disturb;
	bool dnd_id_found = false;
	GVariant *body = NULL;
	GList *dnd_found_list = NULL;
	dnd_alarm_id_s *dnd_id_data = NULL;
	uid_t uid;

	dnd_found_list = g_list_first(_dnd_alarm_id_list);

	for (; dnd_found_list != NULL; dnd_found_list = dnd_found_list->next) {
		dnd_id_data = dnd_found_list->data;
		uid = dnd_id_data->uid;

		ret = noti_system_setting_get_do_not_disturb(&do_not_disturb, uid);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("get do not disturb failed");
			return ret;
		}

		if (alarm_id == dnd_id_data->dnd_start_id && !do_not_disturb) {
			do_not_disturb = 1;
			dnd_id_found = true;
			break;
		} else if (alarm_id == dnd_id_data->dnd_end_id && do_not_disturb) {
			do_not_disturb = 0;
			dnd_id_found = true;
			break;
		}
	}

	if (dnd_id_found == false) {
		ERR("notification wrong alarm [%d]", alarm_id);
		return -1;
	}

	notification_setting_db_update_do_not_disturb(do_not_disturb, uid);

	body = g_variant_new("(ii)", do_not_disturb, uid);

	ret = send_notify(body, "change_dnd_notify", &_monitoring_hash, PROVIDER_NOTI_INTERFACE_NAME, uid);
	g_variant_unref(body);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to send notify [%d]", ret);
		return ret;
	}

	INFO("dnd[%d], uid[%d]", do_not_disturb, uid);
	return ret;
}

static int _get_current_time(struct tm *date)
{
	time_t now;

	if (date == NULL) {
		ERR("Invalid tm data");
		return NOTIFICATION_ERROR_INVALID_PARAMETER;
	}

	time(&now);
	localtime_r(&now, date);

	return NOTIFICATION_ERROR_NONE;
}

static int _noti_system_setting_set_alarm(int week_flag, int hour, int min, alarm_cb_t handler, alarm_id_t *dnd_schedule_alarm_id)
{
	int err = NOTIFICATION_ERROR_NONE;
	struct tm struct_time;
	alarm_entry_t *alarm_info = NULL;
	alarm_date_t alarm_time;
	alarm_id_t alarm_id = -1;

	err = alarmmgr_init("notification");
	if (err < 0) {
		ERR("alarmmgr_init failed [%d]", err);
		goto out;
	}

	err = alarmmgr_set_cb(handler, NULL);
	if (err < 0) {
		ERR("alarmmgr_set_cb failed [%d]", err);
		goto out;
	}

	err = _get_current_time(&struct_time);
	if (err != NOTIFICATION_ERROR_NONE) {
		ERR("get_current_time failed [%d]", err);
		goto out;
	}

	alarm_info = alarmmgr_create_alarm();
	if (alarm_info == NULL) {
		ERR("alarmmgr_create_alarm failed");
		goto out;
	}

	alarm_time.year = struct_time.tm_year + 1900;
	alarm_time.month = struct_time.tm_mon + 1;
	alarm_time.day = struct_time.tm_mday;
	alarm_time.hour = hour;
	alarm_time.min = min;
	alarm_time.sec = 0;

	err = alarmmgr_set_time(alarm_info, alarm_time);
	if (err != ALARMMGR_RESULT_SUCCESS) {
		ERR("alarmmgr_set_time failed [%d]", err);
		goto out;
	}

	if (week_flag) {
		err = alarmmgr_set_repeat_mode(alarm_info, ALARM_REPEAT_MODE_WEEKLY, week_flag);
		if (err != ALARMMGR_RESULT_SUCCESS) {
			ERR("alarmmgr_set_repeat_mode failed [%d]", err);
			goto out;
		}
	}

	err = alarmmgr_set_type(alarm_info, ALARM_TYPE_VOLATILE);
	if (err != ALARMMGR_RESULT_SUCCESS) {
		ERR("alarmmgr_set_type failed [%d]", err);
		goto out;
	}

	err = alarmmgr_add_alarm_with_localtime(alarm_info, NULL, &alarm_id);
	if (err != ALARMMGR_RESULT_SUCCESS) {
		ERR("alarmmgr_add_alarm_with_localtime failed [%d]", err);
		goto out;
	}

	*dnd_schedule_alarm_id = alarm_id;

out:
	if (alarm_info)
		alarmmgr_free_alarm(alarm_info);

	DBG("alarm_id [%d]", alarm_id);
	return err;
}

static gint _dnd_data_compare(gconstpointer a, gconstpointer b)
{
	const dnd_alarm_id_s *data = NULL;

	if (!a)
		return -1;
	data = a;
	if (data->uid == GPOINTER_TO_UINT(b))
		return 0;

	return 1;
}

static dnd_alarm_id_s *__malloc_dnd_alarm_id_s(uid_t uid, alarm_id_t start_alarm_id, alarm_id_t end_alarm_id)
{
	dnd_alarm_id_s *dnd_id_data = NULL;

	dnd_id_data = (dnd_alarm_id_s *)malloc(sizeof(dnd_alarm_id_s));
	if (dnd_id_data == NULL) {
		ERR("Failed to alloc memory");
		return NULL;
	}

	dnd_id_data->uid = uid;
	dnd_id_data->dnd_start_id = start_alarm_id;
	dnd_id_data->dnd_end_id = end_alarm_id;

	return dnd_id_data;
}

static int _add_alarm(int dnd_schedule_day, int dnd_start_hour, int dnd_start_min, int dnd_end_hour, int dnd_end_min, uid_t uid)
{
	int ret = NOTIFICATION_ERROR_NONE;
	GList *found_list = NULL;
	dnd_alarm_id_s *dnd_id_data = NULL;
	alarm_id_t dnd_schedule_start_alarm_id;
	alarm_id_t dnd_schedule_end_alarm_id;

	ret = _noti_system_setting_set_alarm(dnd_schedule_day,
				dnd_start_hour, dnd_start_min,
				_dnd_schedule_alarm_cb,
				&dnd_schedule_start_alarm_id);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to set start alarm [%d]", ret);
		return ret;
	}

	if (dnd_end_hour < dnd_start_hour ||
	    (dnd_end_hour == dnd_start_hour && dnd_end_min <= dnd_start_min)) {
		dnd_schedule_day = (dnd_schedule_day << 1);
		if (dnd_schedule_day & (DND_SCHEDULE_WEEK_FLAG_SATURDAY << 1))
			dnd_schedule_day = (dnd_schedule_day & DND_SCHEDULE_WEEK_FLAG_ALL)
					| DND_SCHEDULE_WEEK_FLAG_SUNDAY;
	}

	ret = _noti_system_setting_set_alarm(dnd_schedule_day,
				dnd_end_hour, dnd_end_min,
				_dnd_schedule_alarm_cb,
				&dnd_schedule_end_alarm_id);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to set end alarm [%d]", ret);
		return ret;
	}

	if (_dnd_alarm_id_list == NULL) {
		dnd_id_data = __malloc_dnd_alarm_id_s(uid, dnd_schedule_start_alarm_id, dnd_schedule_end_alarm_id);
		if (!dnd_id_data)
			return NOTIFICATION_ERROR_OUT_OF_MEMORY;
		_dnd_alarm_id_list = g_list_append(_dnd_alarm_id_list, dnd_id_data);
	} else {
		found_list = g_list_find_custom(g_list_first(_dnd_alarm_id_list), GUINT_TO_POINTER(uid),
						(GCompareFunc)_dnd_data_compare);
		if (found_list) {
			dnd_id_data = g_list_nth_data(found_list, 0);
			if (dnd_id_data->dnd_start_id) {
				alarmmgr_remove_alarm(dnd_id_data->dnd_start_id);
				dnd_id_data->dnd_start_id = dnd_schedule_start_alarm_id;
			}
			if (dnd_id_data->dnd_end_id) {
				alarmmgr_remove_alarm(dnd_id_data->dnd_end_id);
				dnd_id_data->dnd_end_id = dnd_schedule_end_alarm_id;
			}
		} else {
			dnd_id_data = __malloc_dnd_alarm_id_s(uid, dnd_schedule_start_alarm_id, dnd_schedule_end_alarm_id);
			if (!dnd_id_data)
				return NOTIFICATION_ERROR_OUT_OF_MEMORY;
			_dnd_alarm_id_list = g_list_append(_dnd_alarm_id_list, dnd_id_data);
		}
	}

	return ret;
}

static void _delete_alarm(uid_t uid)
{
	GList *found_list = NULL;
	dnd_alarm_id_s *dnd_id_data = NULL;

	if (_dnd_alarm_id_list == NULL)
		return;

	_dnd_alarm_id_list = g_list_first(_dnd_alarm_id_list);
	found_list = g_list_find_custom(_dnd_alarm_id_list, GUINT_TO_POINTER(uid),
					(GCompareFunc)_dnd_data_compare);

	if (found_list == NULL)
		return;

	dnd_id_data = g_list_nth_data(found_list, 0);

	if (dnd_id_data->dnd_start_id)
		alarmmgr_remove_alarm(dnd_id_data->dnd_start_id);
	if (dnd_id_data->dnd_end_id)
		alarmmgr_remove_alarm(dnd_id_data->dnd_end_id);

	_dnd_alarm_id_list = g_list_remove(g_list_first(_dnd_alarm_id_list), dnd_id_data);
	free(dnd_id_data);
}

/* update_noti_sys_setting */
int notification_update_noti_sys_setting(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	int pre_do_not_disturb = 0;
	int do_not_disturb = 0;
	int visivility_class = 0;
	int dnd_schedule_enabled = 0;
	int dnd_schedule_day = 0;
	int dnd_start_hour = 0;
	int dnd_start_min = 0;
	int dnd_end_hour = 0;
	int dnd_end_min = 0;
	int lock_screen_level = 0;
	uid_t param_uid;
	GVariant *body = NULL;

	g_variant_get(parameters, "(iiiiiiiiii)",
				&do_not_disturb,
				&visivility_class,
				&dnd_schedule_enabled,
				&dnd_schedule_day,
				&dnd_start_hour,
				&dnd_start_min,
				&dnd_end_hour,
				&dnd_end_min,
				&lock_screen_level,
				&param_uid);

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = noti_system_setting_get_do_not_disturb(&pre_do_not_disturb, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to get do not disturb");
		return ret;
	}

	ret = notification_setting_db_update_system_setting(do_not_disturb,
				visivility_class,
				dnd_schedule_enabled,
				dnd_schedule_day,
				dnd_start_hour,
				dnd_start_min,
				dnd_end_hour,
				dnd_end_min,
				lock_screen_level,
				param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to setting db update system setting : %d", ret);
		return ret;
	}

	if (pre_do_not_disturb != do_not_disturb) {
		body = g_variant_new("(ii)", do_not_disturb, param_uid);
		ret = send_notify(body, "change_dnd_notify", &_monitoring_hash, PROVIDER_NOTI_INTERFACE_NAME, param_uid);
		g_variant_unref(body);
	}

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	if (dnd_schedule_enabled) {
		ret = _add_alarm(dnd_schedule_day, dnd_start_hour, dnd_start_min,
				dnd_end_hour, dnd_end_min, param_uid);
		if (ret != NOTIFICATION_ERROR_NONE)
			ERR("Failed to add alarm for dnd_schedule");
	} else {
		_delete_alarm(param_uid);
	}

	INFO("do_not_disturb [%d] visivility_class [%d] set_schedule [%d] lock_screen_level [%d]",
			do_not_disturb, visivility_class, dnd_schedule_enabled, lock_screen_level);
	return ret;
}

int notification_add_noti_template(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	notification_h noti;
	int ret;
	GVariant *body = NULL;
	GVariant *coupled_body = NULL;
	char *template_name = NULL;
	int count = 0;

	noti = notification_create(NOTIFICATION_TYPE_NOTI);
	if (noti != NULL) {
		g_variant_get(parameters, "(v&s)", &coupled_body, &template_name);
		g_variant_get(coupled_body, "(v)", &body);

		ret = notification_ipc_make_noti_from_gvariant(noti, body);
		g_variant_unref(coupled_body);
		g_variant_unref(body);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to make a notification [%d]", ret);
			goto out;
		}

		ret = notification_noti_check_count_for_template(noti, &count);
		if (count >= NOTI_TEMPLATE_LIMIT) {
			ERR("Exceed limit value that is saved tamplate  [%d]", ret);
			ret = NOTIFICATION_ERROR_MAX_EXCEEDED;
			goto out;
		}

		ret = notification_noti_add_template(noti, template_name);
		if (ret != NOTIFICATION_ERROR_NONE)
			ERR("Failed to add a notification [%d]", ret);
	} else {
		ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

out:
	if (noti)
		notification_free(noti);

	return ret;
}

int notification_get_noti_template(GVariant *parameters, GVariant **reply_body, pid_t pid, uid_t uid)
{
	int ret;
	char app_id[BUF_LEN] = {0, };
	char *template_name = NULL;
	notification_h noti;

	noti = notification_create(NOTIFICATION_TYPE_NOTI);
	if (noti != NULL) {
		g_variant_get(parameters, "(&s)", &template_name);
		INFO("template name[%s]", template_name);

		ret = aul_app_get_appid_bypid_for_uid(pid, app_id, sizeof(app_id), uid);
		if (ret != AUL_R_OK) {
			ERR("Failed to get app_id [%d]", ret);
			ret = NOTIFICATION_ERROR_INVALID_PARAMETER;
			goto out;
		}

		ret = notification_noti_get_package_template(noti, app_id, template_name);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to get template [%d]", ret);
			goto out;
		}

		*reply_body = notification_ipc_make_gvariant_from_noti(noti, false);
		if (*reply_body == NULL) {
			ERR("Failed to make reply_body");
			ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
		}
	} else {
		ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

out:
	if (noti)
		notification_free(noti);

	return ret;
}

int notification_get_noti_package_template(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	char *app_id = NULL;
	char *template_name = NULL;
	notification_h noti;

	noti = notification_create(NOTIFICATION_TYPE_NOTI);
	if (noti != NULL) {
		g_variant_get(parameters, "(&s&s)", &app_id, &template_name);
		INFO("app_id[%s] template name[%s]", app_id, template_name);
		ret = notification_noti_get_package_template(noti, app_id, template_name);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to get template [%d]", ret);
			goto out;
		}

		*reply_body = notification_ipc_make_gvariant_from_noti(noti, false);
		if (*reply_body == NULL) {
			ERR("Failed to make reply_body");
			ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
		}
	} else {
		ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

out:
	if (noti)
		notification_free(noti);

	return ret;
}

int notification_get_block_state(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	int dnd;
	int dnd_except;
	int allow_to_notify;
	char *app_id;
	uid_t param_uid;
	GVariant *body = NULL;

	g_variant_get(parameters, "(&si)", &app_id, &param_uid);

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("_validate_uid fail [%d]", ret);
		return NOTIFICATION_ERROR_IO_ERROR;
	}

	ret = notification_get_dnd_and_allow_to_notify(app_id, &dnd, &dnd_except, &allow_to_notify, param_uid);
	INFO("app_id[%s] dnd[%d] dnd_except[%d] allow_to_notify[%d]",
		app_id, dnd, dnd_except, allow_to_notify);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to get_dnd_and_allow_to_notify [%d]", ret);
		return ret;
	}

	body = g_variant_new("(iii)", dnd, dnd_except, allow_to_notify);
	*reply_body = g_variant_new("(v)", body);

	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	return ret;
}
int notification_load_dnd_allow_exception(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	int count = 0;
	int i;
	uid_t param_uid;
	GVariant *body = NULL;
	GVariantBuilder *builder;
	dnd_allow_exception_h dnd_allow_exception = NULL;
	dnd_allow_exception_h temp;

	g_variant_get(parameters, "(i)", &param_uid);

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = notification_system_setting_load_dnd_allow_exception(&dnd_allow_exception, &count, param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to load dnd_allow_exception [%d]", ret);
		return ret;
	}

	builder = g_variant_builder_new(G_VARIANT_TYPE("a(v)"));

	if (dnd_allow_exception) {
		for (i = 0; i < count; i++) {
			temp = dnd_allow_exception + i;
			body = notification_ipc_make_gvariant_from_dnd_allow_exception(temp);
			g_variant_builder_add(builder, "(v)", body);
		}
		free(dnd_allow_exception);
	}

	*reply_body = g_variant_new("(ia(v))", count, builder);
	g_variant_builder_unref(builder);

	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	return ret;
}

int notification_update_dnd_allow_exception(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int ret;
	int type;
	int value;
	uid_t param_uid;

	g_variant_get(parameters, "(iii)", &type, &value, &param_uid);

	ret = _validate_and_set_param_uid_with_uid(uid, &param_uid);
	if (ret != NOTIFICATION_ERROR_NONE)
		return ret;

	ret = notification_system_setting_update_dnd_allow_exception(type,
			value,
			param_uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to update dnd_allow_exception [%d]", ret);
		return ret;
	}

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make reply_body");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	return ret;
}

static event_sender_info_s *__find_sender_info_by_priv_id(int priv_id)
{
	event_sender_info_s *event_info;
	GList *find_list;

	if (__event_list == NULL)
		return NULL;

	__event_list = g_list_first(__event_list);
	find_list = g_list_find_custom(__event_list, GINT_TO_POINTER(priv_id),
				       (GCompareFunc)__priv_id_compare);

	if (find_list)
		event_info = g_list_nth_data(find_list, 0);
	else
		return NULL;

	return event_info;
}

int notification_send_noti_event(GVariant *parameters, GVariant **reply_body)
{
	int ret;
	int event_type;
	int priv_id;
	event_sender_info_s *info;
	GVariant *coupled_body = NULL;
	GVariant *body = NULL;
	notification_h noti;

	noti = notification_create(NOTIFICATION_TYPE_NOTI);
	if (noti != NULL) {
		g_variant_get(parameters, "(vi)", &coupled_body, &event_type);
		g_variant_get(coupled_body, "(v)", &body);

		ret = notification_ipc_make_noti_from_gvariant(noti, body);
		g_variant_unref(coupled_body);
		g_variant_unref(body);

		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to make a notification from gvariant");
			goto out;
		}

		ret = notification_get_id(noti, NULL, &priv_id);
		if (ret != NOTIFICATION_ERROR_NONE)
			goto out;

		info = __find_sender_info_by_priv_id(priv_id);
		if (info == NULL || info->busname == NULL) {
			ret = NOTIFICATION_ERROR_INVALID_PARAMETER;
			goto out;
		}

		if (!info->watcher_id) {
			if (!is_existed_busname(info->busname)) {
				__event_list = g_list_remove(g_list_first(__event_list), info);
				__free_event_info(info);
				ret = NOTIFICATION_ERROR_IO_ERROR;
				goto out;
			} else {
				info->watcher_id = __insert_sender_watcher_id(info);
			}
		}

		ret = send_event_notify_by_busname(parameters, "send_event", info->busname, PROVIDER_NOTI_EVENT_INTERFACE_NAME);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to send event [%d]", ret);
			goto out;
		}

		*reply_body = g_variant_new("()");
		if (*reply_body == NULL) {
			ERR("Failed to make reply body");
			ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
		}
	} else {
		ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

out:
	if (noti)
		notification_free(noti);

	return ret;
}

int notification_send_noti_event_by_priv_id(GVariant *parameters, GVariant **reply_body)
{
	int ret;
	int event_type;
	int priv_id;
	bool event_flag = false;
	notification_h noti;
	GVariant *body = NULL;
	event_sender_info_s *info;

	noti = notification_create(NOTIFICATION_TYPE_NOTI);
	if (noti != NULL) {
		g_variant_get(parameters, "(ii)", &priv_id, &event_type);

		ret = notification_noti_get_by_priv_id(noti, priv_id);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to get notification by priv id [%d]", ret);
			goto out;
		}

		ret = notification_get_event_flag(noti, &event_flag);
		if (ret != NOTIFICATION_ERROR_NONE || event_flag == false) {
			ret = NOTIFICATION_ERROR_INVALID_PARAMETER;
			goto out;
		}

		info = __find_sender_info_by_priv_id(priv_id);
		if (info == NULL || info->busname == NULL) {
			ret = NOTIFICATION_ERROR_INVALID_PARAMETER;
			ERR("No sender info by priv_id[%d]", priv_id);
			goto out;
		}

		if (!info->watcher_id) {
			if (!is_existed_busname(info->busname)) {
				__event_list = g_list_remove(g_list_first(__event_list), info);
				__free_event_info(info);
				ret = NOTIFICATION_ERROR_IO_ERROR;
				goto out;
			} else {
				info->watcher_id = __insert_sender_watcher_id(info);
			}
		}

		body = notification_ipc_make_gvariant_from_noti(noti, false);
		if (body == NULL) {
			ERR("Can't make gvariant to noti");
			ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
			goto out;
		}

		ret = send_event_notify_by_busname(g_variant_new("(vi)", body, event_type), "send_event", info->busname, PROVIDER_NOTI_EVENT_INTERFACE_NAME);
		*reply_body = g_variant_new("()");
		if (*reply_body == NULL) {
			ERR("Failed to make reply body");
			ret = NOTIFICATION_ERROR_OUT_OF_MEMORY;
		}
	} else {
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

out:
	if (noti)
		notification_free(noti);

	return ret;
}

int notification_check_event_receiver(GVariant *parameters, GVariant **reply_body)
{
	int priv_id;
	bool available = false;
	GList *find_list = NULL;
	event_sender_info_s *info = NULL;

	g_variant_get(parameters, "(i)", &priv_id);
	INFO("priv_id[%d]", priv_id);

	__event_list = g_list_first(__event_list);
	find_list = g_list_find_custom(__event_list, GINT_TO_POINTER(priv_id),
				       (GCompareFunc)__priv_id_compare);
	if (find_list) {
		info = g_list_nth_data(find_list, 0);
		if (!info->watcher_id) {
			if (!is_existed_busname(info->busname)) {
				__event_list = g_list_remove(g_list_first(__event_list), info);
				__free_event_info(info);
			} else {
				available = true;
				info->watcher_id = __insert_sender_watcher_id(info);
			}
		} else {
			available = true;
		}
	}

	*reply_body = g_variant_new("(i)", (int)available);
	if (*reply_body == NULL) {
		ERR("Failed to make gvariant to noti");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	INFO("Receiver available[%d]", available);
	return NOTIFICATION_ERROR_NONE;
}

int notification_reset_event_receiver(GVariant *parameters, GVariant **reply_body, const char *sender)
{
	int priv_id;

	g_variant_get(parameters, "(i)", &priv_id);
	INFO("priv_id[%d]", priv_id);

	__add_sender_info(priv_id, sender);

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make gvariant to noti");
		return NOTIFICATION_ERROR_OUT_OF_MEMORY;
	}

	return NOTIFICATION_ERROR_NONE;
}

HAPI int notification_delete_noti_by_app_id(const char *app_id, uid_t uid)
{
	GVariant *deleted_noti_list;
	GVariantBuilder *builder;
	int num_deleted = 0;
	int *list_deleted = NULL;
	int ret;
	int i;

	ret = notification_noti_delete_all(NOTIFICATION_TYPE_NONE, app_id, &num_deleted, &list_deleted, uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to delete notifications [%d]", ret);
		goto out;
	}

	if (num_deleted > 0) {
		builder = g_variant_builder_new(G_VARIANT_TYPE("a(i)"));

		for (i = 0; i < num_deleted; i++)
			g_variant_builder_add(builder, "(i)", *(list_deleted + i));

		deleted_noti_list = g_variant_new("(a(i)ii)", builder, num_deleted, uid);
		ret = send_notify(deleted_noti_list, "delete_multiple_notify", &_monitoring_hash, PROVIDER_NOTI_INTERFACE_NAME, uid);
		g_variant_builder_unref(builder);
		g_variant_unref(deleted_noti_list);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("Failed to send notify [%d]", ret);
			goto out;
		}

		for (i = 0; i < num_deleted; i++) {
			ret = __delete_sender_info(*(list_deleted + i));
			if (ret != NOTIFICATION_ERROR_NONE)
				goto out;
		}
	}
out:
	if (list_deleted != NULL)
		free(list_deleted);
	return ret;
}

static int _check_dnd_schedule(uid_t uid)
{
	int ret;
	notification_system_setting_h setting = NULL;
	bool dnd_schedule_enabled = false;
	int dnd_schedule_day = 0;
	int dnd_start_hour = 0;
	int dnd_start_min = 0;
	int dnd_end_hour = 0;
	int dnd_end_min = 0;

	ret = noti_system_setting_load_system_setting(&setting, uid);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("noti_system_setting_load_system_setting fail [%d]", ret);
		return ret;
	}

	ret = notification_system_setting_dnd_schedule_get_enabled(setting,
				&dnd_schedule_enabled);
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("system_setting_dnd_schedule_get_enabled fail [%d]", ret);
		goto out;
	}

	if (dnd_schedule_enabled) {
		ret = notification_system_setting_dnd_schedule_get_day(setting,
				&dnd_schedule_day);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("system_setting_dnd_schedule_get_day fail [%d]", ret);
			goto out;
		}

		ret = notification_system_setting_dnd_schedule_get_start_time(setting,
				&dnd_start_hour, &dnd_start_min);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("system_setting_dnd_schedule_get_start_time fail [%d]", ret);
			goto out;
		}

		ret = notification_system_setting_dnd_schedule_get_end_time(setting,
				&dnd_end_hour, &dnd_end_min);
		if (ret != NOTIFICATION_ERROR_NONE) {
			ERR("system_setting_dnd_schedule_get_end_time fail [%d]", ret);
			goto out;
		}

		_add_alarm(dnd_schedule_day, dnd_start_hour, dnd_start_min, dnd_end_hour, dnd_end_min, uid);
	}

out:
	notification_system_setting_free_system_setting(setting);

	return ret;
}

static gboolean __refresh_setting_table(gpointer data)
{
	notification_setting_refresh_setting_table(
		tzplatform_getuid(TZ_SYS_DEFAULT_USER));
	return G_SOURCE_REMOVE;
}

/*!
 * MAIN THREAD
 * Do not try to do any other operation in these functions
 */
HAPI int notification_service_init(int restart_count)
{
	int ret, i;
	int count = 0;
	uid_t *uids = NULL;

	_monitoring_hash = g_hash_table_new(g_direct_hash, g_direct_equal);
	ret = notification_db_init();
	if (ret != NOTIFICATION_ERROR_NONE) {
		ERR("Failed to init DB [%d]", ret);
		return ret;
	}

	ret = notification_register_dbus_interface();
	if (ret != SERVICE_COMMON_ERROR_NONE) {
		ERR("Failed to register dbus interface [%d]", ret);
		return NOTIFICATION_ERROR_IO_ERROR;
	}

	notification_init_default_viewer();

	if (restart_count == 0)
		notification_noti_init_data();

	notification_system_setting_init_system_setting_table(tzplatform_getuid(TZ_SYS_DEFAULT_USER));

	ret = notification_system_setting_get_dnd_schedule_enabled_uid(&uids, &count);
	if (ret == NOTIFICATION_ERROR_NONE && count > 0) {
		for (i = 0; i < count; i++)
			_check_dnd_schedule(uids[i]);
	}
	if (uids)
		free(uids);

	g_idle_add(__refresh_setting_table, NULL);

	INFO("Successfully initialized");
	return NOTIFICATION_ERROR_NONE;
}

HAPI int notification_service_fini(void)
{
	pkgmgr_fini();
	DBG("Successfully Finalized");
	return NOTIFICATION_ERROR_NONE;
}

/* End of a file */
