/*
 * Copyright 2016  Samsung Electronics Co., Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>

#include <dlog.h>
#include <package-manager.h>

#include "service_common.h"
#include "debug.h"
#include "pkgmgr.h"
#include "conf.h"

#define PKGMGR_EVENT_MIN PKGMGR_EVENT_DOWNLOAD
#define PKGMGR_EVENT_MAX PKGMGR_EVENT_APP_DISABLE

struct item {
	uid_t uid;
	char *pkgname;
	char *icon;

	enum pkgmgr_event_type type;
	enum pkgmgr_status status;
};

struct event_item {
	int (*cb)(uid_t uid, const char *pkgname, enum pkgmgr_status status, double value, void *data);
	void *data;
};

static struct {
	pkgmgr_client *listen_pc;
	GList *item_list;

} s_info = {
	.listen_pc = NULL,
	.item_list = NULL,
};

static GList *event_list[PKGMGR_EVENT_MAX + 1];

static inline void invoke_callback(uid_t uid, const char *pkgname, struct item *item, double value)
{
	GList *l;
	struct event_item *e_item;

	for (l = event_list[item->type]; l != NULL; l = g_list_next(l)) {
		e_item = (struct event_item *)g_list_nth_data(l, 0);
		if (e_item->cb)
			e_item->cb(uid, pkgname, item->status, value, e_item->data);
	}
}

static inline int is_valid_status(struct item *item, const char *status)
{
	const char *expected_status;

	switch (item->type) {
	case PKGMGR_EVENT_DOWNLOAD:
		expected_status = "download";
		break;
	case PKGMGR_EVENT_UNINSTALL:
		expected_status = "uninstall";
		break;
	case PKGMGR_EVENT_INSTALL:
		expected_status = "install";
		break;
	case PKGMGR_EVENT_UPDATE:
		expected_status = "update";
		break;
	case PKGMGR_EVENT_RECOVER:
		expected_status = "recover";
		break;
	case PKGMGR_EVENT_APP_ENABLE:
		expected_status = "app_enable_app";
		break;
	case PKGMGR_EVENT_APP_DISABLE:
		expected_status = "app_disable_app";
		break;
	default:
		return 0;
	}

	return !strcasecmp(status, expected_status);
}

static struct item *find_item(const char *pkgname, uid_t uid)
{
	GList *l;
	struct item *item;

	if (!pkgname) {
		ERR("Invalid pkgname");
		return NULL;
	}

	for (l = s_info.item_list; l != NULL; l = g_list_next(l)) {
		item = g_list_nth_data(l, 0);
		if (uid != item->uid || strcmp(item->pkgname, pkgname))
			continue;

		return item;
	}

	DBG("Package [%s] is not found", pkgname);
	return NULL;
}

static int start_cb(uid_t uid, const char *pkgname, const char *val, void *data)
{
	struct item *item;

	INFO("pkgname[%s] val[%s]", pkgname, val);

	item = calloc(1, sizeof(*item));
	if (!item) {
		ERR("calloc [%d]", errno);
		return SERVICE_COMMON_ERROR_OUT_OF_MEMORY;
	}

	item->pkgname = strdup(pkgname);
	if (!item->pkgname) {
		ERR("strdup [%d]", errno);
		free(item);
		return SERVICE_COMMON_ERROR_OUT_OF_MEMORY;
	}

	item->uid = uid;
	item->status = PKGMGR_STATUS_START;

	if (!strcasecmp(val, "download")) {
		item->type = PKGMGR_EVENT_DOWNLOAD;
	} else if (!strcasecmp(val, "uninstall")) {
		item->type = PKGMGR_EVENT_UNINSTALL;
	} else if (!strcasecmp(val, "install")) {
		item->type = PKGMGR_EVENT_INSTALL;
	} else if (!strcasecmp(val, "update")) {
		item->type = PKGMGR_EVENT_UPDATE;
	} else if (!strcasecmp(val, "recover")) {
		item->type = PKGMGR_EVENT_RECOVER;
	} else if (!strcasecmp(val, "enable_app")) {
		item->type = PKGMGR_EVENT_APP_ENABLE;
	} else if (!strcasecmp(val, "disable_app")) {
		item->type = PKGMGR_EVENT_APP_DISABLE;
	} else {
		free(item->pkgname);
		free(item);
		ERR("Invalid val [%s]", val);
		return SERVICE_COMMON_ERROR_INVALID_PARAMETER;
	}
	s_info.item_list = g_list_append(s_info.item_list, item);

	invoke_callback(uid, pkgname, item, 0.0f);
	return SERVICE_COMMON_ERROR_NONE;
}

static int icon_path_cb(uid_t uid, const char *pkgname, const char *val, void *data)
{
	struct item *item;

	DBG("pkgname[%s] val[%s]", pkgname, val);

	item = find_item(pkgname, uid);
	if (!item)
		return SERVICE_COMMON_ERROR_NOT_EXIST;

	if (item->icon)
		free(item->icon);

	item->icon = strdup(val);
	if (!item->icon) {
		ERR("strdup [%d]", errno);
		return SERVICE_COMMON_ERROR_OUT_OF_MEMORY;
	}

	return SERVICE_COMMON_ERROR_NONE;
}

static int command_cb(uid_t uid, const char *pkgname, const char *val, void *data)
{
	struct item *item;

	DBG("pkgname[%s] val[%s]", pkgname, val);

	item = find_item(pkgname, uid);
	if (!item)
		return SERVICE_COMMON_ERROR_NOT_EXIST;

	if (!is_valid_status(item, val)) {
		DBG("Invalid status[%d] val[%s]", item->type, val);
		return SERVICE_COMMON_ERROR_INVALID_PARAMETER;
	}

	item->status = PKGMGR_STATUS_COMMAND;
	invoke_callback(uid, pkgname, item, 0.0f);
	return SERVICE_COMMON_ERROR_NONE;
}

static int error_cb(uid_t uid, const char *pkgname, const char *val, void *data)
{
	/* val = error */
	struct item *item;

	INFO("pkgname[%s] val[%s]", pkgname, val);

	item = find_item(pkgname, uid);
	if (!item)
		return SERVICE_COMMON_ERROR_NOT_EXIST;

	item->status = PKGMGR_STATUS_ERROR;
	invoke_callback(uid, pkgname, item, 0.0f);
	return SERVICE_COMMON_ERROR_NONE;
}

static int change_pkgname_cb(uid_t uid, const char *pkgname, const char *val, void *data)
{
	struct item *item;
	char *new_pkgname;

	INFO("pkgname[%s] val[%s]", pkgname, val);

	item = find_item(pkgname, uid);
	if (!item)
		return SERVICE_COMMON_ERROR_NOT_EXIST;

	new_pkgname = strdup(val);
	if (!new_pkgname) {
		ERR("strdup: %d", errno);
		return SERVICE_COMMON_ERROR_OUT_OF_MEMORY;
	}

	free(item->pkgname);
	item->pkgname = new_pkgname;
	return SERVICE_COMMON_ERROR_NONE;
}

static int download_cb(uid_t uid, const char *pkgname, const char *val, void *data)
{
	/* val = integer */
	struct item *item;
	double value;

	INFO("pkgname[%s] val[%s]", pkgname, val);

	item = find_item(pkgname, uid);
	if (!item) {
		DBG("ITEM is not started from the start_cb");
		return SERVICE_COMMON_ERROR_INVALID_PARAMETER;
	}

	if (item->type != PKGMGR_EVENT_DOWNLOAD) {
		DBG("TYPE is not \"download\" [%d]", item->type);
		item->type = PKGMGR_EVENT_DOWNLOAD;
	}

	switch (item->status) {
	case PKGMGR_STATUS_START:
	case PKGMGR_STATUS_COMMAND:
		item->status = PKGMGR_STATUS_PROCESSING;
	case PKGMGR_STATUS_PROCESSING:
		break;
	default:
		ERR("Invalid status, pkgname[%s] val[%s]", pkgname, val);
		return SERVICE_COMMON_ERROR_INVALID_PARAMETER;
	}

	if (val) {
		if (sscanf(val, "%lf", &value) != 1)
			value = (double)SERVICE_COMMON_ERROR_INVALID_PARAMETER;
	} else {
		value = (double)SERVICE_COMMON_ERROR_INVALID_PARAMETER;
	}

	invoke_callback(uid, pkgname, item, value);
	return SERVICE_COMMON_ERROR_NONE;
}

static int progress_cb(uid_t uid, const char *pkgname, const char *val, void *data)
{
	/* val = integer */
	struct item *item;
	double value;

	item = find_item(pkgname, uid);
	if (!item) {
		ERR("ITEM is not started from the start_cb");
		return SERVICE_COMMON_ERROR_INVALID_PARAMETER;
	}

	switch (item->status) {
	case PKGMGR_STATUS_START:
	case PKGMGR_STATUS_COMMAND:
		item->status = PKGMGR_STATUS_PROCESSING;
	case PKGMGR_STATUS_PROCESSING:
		break;
	default:
		ERR("Invalid status, pkgname[%s] val[%s]", pkgname, val);
		return SERVICE_COMMON_ERROR_INVALID_PARAMETER;
	}

	if (val) {
		if (sscanf(val, "%lf", &value) != 1)
			value = (double)SERVICE_COMMON_ERROR_INVALID_PARAMETER;
	} else {
		value = (double)SERVICE_COMMON_ERROR_INVALID_PARAMETER;
	}

	invoke_callback(uid, pkgname, item, value);
	return SERVICE_COMMON_ERROR_NONE;
}

static int end_cb(uid_t uid, const char *pkgname, const char *val, void *data)
{
	struct item *item;

	DBG("[%s] %s", pkgname, val);

	item = find_item(pkgname, uid);
	if (!item)
		return SERVICE_COMMON_ERROR_NOT_EXIST;

	item->status = !strcasecmp(val, "ok") ? PKGMGR_STATUS_END : PKGMGR_STATUS_ERROR;

	invoke_callback(uid, pkgname, item, 0.0f);

	s_info.item_list = g_list_remove(s_info.item_list, item);
	free(item->icon);
	free(item->pkgname);
	free(item);
	return SERVICE_COMMON_ERROR_NONE;
}

static struct pkgmgr_handler {
	const char *key;
	int (*func)(uid_t uid, const char *package, const char *val, void *data);
} handler[] = {
	{ "install_percent", progress_cb },
	{ "download_percent", download_cb },
	{ "start", start_cb },
	{ "end", end_cb },
	{ "change_pkg_name", change_pkgname_cb },
	{ "icon_path", icon_path_cb },
	{ "command", command_cb },
	{ "error", error_cb },
	{ NULL, NULL },
};

static int status_cb(uid_t target_uid, int req_id, const char *type,
		const char *pkgname, const char *key, const char *val,
		const void *pmsg, void *data)
{
	register int i;
	int ret;

	for (i = 0; handler[i].key; i++) {
		if (strcasecmp(key, handler[i].key))
			continue;

		ret = handler[i].func(target_uid, pkgname, val, data);
		if (ret < 0) {
			DBG("REQ[%d] UID[%d] pkgname[%s] type[%s] key[%s] val[%s] ret[%d]",
					req_id, target_uid, pkgname, type, key, val, ret);
		}
	}

	return SERVICE_COMMON_ERROR_NONE;
}

static int app_status_cb(uid_t target_uid, int req_id, const char *pkg_type,
		const char *pkgid, const char *appid, const char *key, const char *val,
		const void *pmsg, void *data)
{
	register int i;
	int ret;

	for (i = 0; handler[i].key; i++) {
		if (strcasecmp(key, handler[i].key))
			continue;

		ret = handler[i].func(target_uid, appid, val, data);
		if (ret < 0) {
			DBG("REQ[%d] UID[%d] appid[%s] type[%s] key[%s] val[%s] ret[%d]",
					req_id, target_uid, appid, pkg_type, key, val, ret);
		}
	}

	return SERVICE_COMMON_ERROR_NONE;
}

static void __free_event_data(gpointer data)
{
	struct event_item *item = (struct event_item *)data;

	if (item)
		free(item);
}

static void __free_item_data(gpointer data)
{
	struct item *ctx = (struct item *)data;

	if (ctx) {
		if (ctx->pkgname)
			free(ctx->pkgname);
		if (ctx->icon)
			free(ctx->icon);
		free(ctx);
	}
}

HAPI int pkgmgr_init(void)
{
	if (s_info.listen_pc)
		return SERVICE_COMMON_ERROR_ALREADY_EXIST;

	s_info.listen_pc = pkgmgr_client_new(PC_LISTENING);
	if (!s_info.listen_pc)
		return SERVICE_COMMON_ERROR_FAULT;

	/* the pkgmgr api returns an id greater than 1 */
	if (pkgmgr_client_listen_status(s_info.listen_pc, status_cb, NULL) <= 0)
		return SERVICE_COMMON_ERROR_FAULT;

	if (pkgmgr_client_listen_app_status(s_info.listen_pc, app_status_cb, NULL) <= 0)
		return SERVICE_COMMON_ERROR_FAULT;

	return SERVICE_COMMON_ERROR_NONE;
}

HAPI int pkgmgr_fini(void)
{
	int i;

	if (!s_info.listen_pc)
		return SERVICE_COMMON_ERROR_INVALID_PARAMETER;

	if (pkgmgr_client_free(s_info.listen_pc) != PKGMGR_R_OK)
		return SERVICE_COMMON_ERROR_FAULT;

	s_info.listen_pc = NULL;

	for (i = PKGMGR_EVENT_MIN; i <= PKGMGR_EVENT_MAX; i++)
		g_list_free_full(event_list[i], __free_event_data);

	g_list_free_full(s_info.item_list, __free_item_data);

	return SERVICE_COMMON_ERROR_NONE;
}

HAPI int pkgmgr_add_event_callback(enum pkgmgr_event_type type, int (*cb)(uid_t uid, const char *pkgname, enum pkgmgr_status status, double value, void *data), void *data)
{
	struct event_item *item;

	item = calloc(1, sizeof(*item));
	if (!item) {
		ERR("calloc [%d]", errno);
		return SERVICE_COMMON_ERROR_OUT_OF_MEMORY;
	}

	item->cb = cb;
	item->data = data;

	event_list[type] = g_list_prepend(event_list[type], item);

	return SERVICE_COMMON_ERROR_NONE;
}

HAPI void *pkgmgr_del_event_callback(enum pkgmgr_event_type type, int (*cb)(uid_t uid, const char *pkgname, enum pkgmgr_status status, double value, void *data), void *data)
{
	struct event_item *item;
	GList *l;
	void *cbdata = NULL;

	for (l = event_list[type]; l != NULL; l = g_list_next(l)) {
		item = g_list_nth_data(l, 0);
		if (item->cb == cb && item->data == data) {
			event_list[type] = g_list_remove(event_list[type], item);
			cbdata = item->data;
			free(item);
			break;
		}
	}

	return cbdata;
}

/* End of a file */
