/*
 * Copyright 2016  Samsung Electronics Co., Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <gio/gio.h>
#include <dlog.h>
#include <notification_setting.h>
#include <notification_setting_internal.h>
#include <notification_setting_service.h>
#include <notification_noti.h>
#include <badge_setting.h>
#include <badge_setting_service.h>
#include <badge_db.h>
#include <package-manager.h>
#include <tzplatform_config.h>
#include <pkgmgr-info.h>

#include "debug.h"
#include "pkgmgr.h"
#include "service_common.h"
#include "notification_service.h"
#include "badge_service.h"
#include "shortcut_service.h"

#define DBUS_NAME "org.freedesktop.DBus"
#define DBUS_OBJECT_PATH "/org/freedesktop/DBus"
#define DBUS_INTERFACE_NAME "org.freedesktop.DBus"

#define PROVIDER_BUS_NAME "org.tizen.data_provider_service"
#define PROVIDER_OBJECT_PATH "/org/tizen/data_provider_service"

static GDBusConnection *_gdbus_conn;
static GHashTable *_noti_pkg_privilege_info;
static GHashTable *_badge_pkg_privilege_info;

uid_t get_sender_uid(const char *sender_name)
{
	GDBusMessage *msg = NULL;
	GDBusMessage *reply = NULL;
	GError *err = NULL;
	GVariant *body;
	uid_t uid = 0;

	msg = g_dbus_message_new_method_call(DBUS_NAME, DBUS_OBJECT_PATH,
			DBUS_INTERFACE_NAME, "GetConnectionUnixUser");
	if (!msg) {
		LOGE("Failed to alloc new method call");
		goto out;
	}

	g_dbus_message_set_body(msg, g_variant_new("(s)", sender_name));
	reply = g_dbus_connection_send_message_with_reply_sync(_gdbus_conn, msg,
			G_DBUS_SEND_MESSAGE_FLAGS_NONE, -1, NULL, NULL, &err);

	if (!reply) {
		if (err != NULL) {
			LOGE("Failed to get uid [%s]", err->message);
			g_error_free(err);
		}
		goto out;
	}

	body = g_dbus_message_get_body(reply);
	g_variant_get(body, "(u)", &uid);

out:
	if (msg)
		g_object_unref(msg);
	if (reply)
		g_object_unref(reply);

	return uid;
}

pid_t get_sender_pid(const char *sender_name)
{
	GDBusMessage *msg = NULL;
	GDBusMessage *reply = NULL;
	GError *err = NULL;
	GVariant *body;
	pid_t pid = 0;

	msg = g_dbus_message_new_method_call(DBUS_NAME, DBUS_OBJECT_PATH,
			DBUS_INTERFACE_NAME, "GetConnectionUnixProcessID");
	if (!msg) {
		LOGE("Failed to alloc new method call");
		goto out;
	}

	g_dbus_message_set_body(msg, g_variant_new("(s)", sender_name));
	reply = g_dbus_connection_send_message_with_reply_sync(_gdbus_conn, msg,
			G_DBUS_SEND_MESSAGE_FLAGS_NONE, -1, NULL, NULL, &err);

	if (!reply) {
		if (err != NULL) {
			LOGE("Failed to get uid [%s]", err->message);
			g_error_free(err);
		}
		goto out;
	}

	body = g_dbus_message_get_body(reply);
	g_variant_get(body, "(u)", &pid);

out:
	if (msg)
		g_object_unref(msg);
	if (reply)
		g_object_unref(reply);

	return pid;
}

bool is_existed_busname(const char *sender_name)
{
	GDBusMessage *msg = NULL;
	GDBusMessage *reply = NULL;
	GError *err = NULL;
	GVariant *body;
	bool is_existed = false;

	msg = g_dbus_message_new_method_call(DBUS_NAME, DBUS_OBJECT_PATH,
			DBUS_INTERFACE_NAME, "NameHasOwner");
	if (!msg) {
		LOGE("Failed to alloc new method call");
		goto out;
	}

	g_dbus_message_set_body(msg, g_variant_new("(s)", sender_name));
	reply = g_dbus_connection_send_message_with_reply_sync(_gdbus_conn, msg,
			G_DBUS_SEND_MESSAGE_FLAGS_NONE, -1, NULL, NULL, &err);

	if (!reply) {
		if (err != NULL) {
			LOGE("Failed to get uid [%s]", err->message);
			g_error_free(err);
		}
		goto out;
	}

	body = g_dbus_message_get_body(reply);
	g_variant_get(body, "(b)", &is_existed);

out:
	if (msg)
		g_object_unref(msg);
	if (reply)
		g_object_unref(reply);

	return is_existed;
}

int send_notify(GVariant *body, char *cmd, GHashTable **monitoring_hash, char *interface_name, uid_t uid)
{
	GError *err = NULL;
	GList *monitoring_list = NULL;
	GList *target_list;
	char *target_bus_name;
	int monitoring_count = 0;
	bool is_existed = false;

	monitoring_list = (GList *)g_hash_table_lookup(*monitoring_hash, GUINT_TO_POINTER(uid));
	target_list = g_list_first(monitoring_list);
	for (; target_list != NULL; ) {
		err = NULL;
		target_bus_name = target_list->data;
		target_list = target_list->next;

		if (g_variant_is_floating(body))
			g_variant_ref(body);

		if (g_dbus_connection_emit_signal(_gdbus_conn,
					target_bus_name,
					PROVIDER_OBJECT_PATH,
					interface_name,
					cmd,
					body,
					&err) == FALSE) {
			if (err != NULL) {
				ERR("Emit signal err [%s]", err->message);
				g_error_free(err);
			}
			is_existed = is_existed_busname(target_bus_name);
			if (is_existed == false)
				delete_monitoring_list(monitoring_hash, target_bus_name, uid);
			ERR("Fail, emit signal to [%s]", target_bus_name);
		}
		monitoring_count++;
		DBG("Success, emit signal to [%s]", target_bus_name);
	}

	DBG("Success, cmd[%s] monitoring count[%d]", cmd, monitoring_count);
	return SERVICE_COMMON_ERROR_NONE;
}

int send_event_notify_by_busname(GVariant *body, char *cmd, char *busname, char *interface_name)
{
	GError *err = NULL;

	if (g_variant_is_floating(body))
		g_variant_ref(body);

	if (g_dbus_connection_emit_signal(_gdbus_conn,
					  busname,
					  PROVIDER_OBJECT_PATH,
					  interface_name,
					  cmd,
					  body,
					  &err) == FALSE) {
		if (err != NULL) {
			ERR("Emit signal err [%s]",
					err->message);
			g_error_free(err);
		}
		ERR("Failed to emit signal to [%s]", busname);
		return SERVICE_COMMON_ERROR_IO_ERROR;
	}
	DBG("Success, Emit signal to [%s] cmd[%s]", busname, cmd);
	return SERVICE_COMMON_ERROR_NONE;
}

/* register service */

static int _monitoring_app_list_compare_cb(gconstpointer a, gconstpointer b)
{
	return strcmp(a, b);
}

int service_register(GVariant *parameters, GVariant **reply_body, const gchar *sender,
		GBusNameAppearedCallback name_appeared_handler,
		GBusNameVanishedCallback name_vanished_handler,
		GHashTable **monitoring_hash,
		uid_t uid)
{
	GList *added_list = NULL;
	const char *bus_name = sender;
	monitoring_info_s *m_info = NULL;
	uid_t request_uid = 0;
	GList *monitoring_list = NULL;

	if (sender == NULL)
		return SERVICE_COMMON_ERROR_IO_ERROR;

	g_variant_get(parameters, "(i)", &request_uid);
	if (uid > NORMAL_UID_BASE && uid != request_uid)
		return SERVICE_COMMON_ERROR_IO_ERROR;

	DBG("service_register : uid %d , request_uid %d", uid, request_uid);
	monitoring_list = (GList *)g_hash_table_lookup(*monitoring_hash, GUINT_TO_POINTER(request_uid));
	added_list = g_list_find_custom(monitoring_list, bus_name,
			(GCompareFunc)_monitoring_app_list_compare_cb);

	if (added_list == NULL) {
		DBG("add new sender to list");
		m_info = (monitoring_info_s *)calloc(1, sizeof(monitoring_info_s));
		if (m_info == NULL) {
			ERR("Failed to alloc memory");
			return SERVICE_COMMON_ERROR_IO_ERROR;
		}

		m_info->bus_name = strdup(bus_name);
		m_info->uid = request_uid;
		m_info->watcher_id = g_bus_watch_name_on_connection(
				_gdbus_conn,
				bus_name,
				G_BUS_NAME_WATCHER_FLAGS_NONE,
				name_appeared_handler,
				name_vanished_handler,
				m_info,
				NULL);
		if (m_info->watcher_id == 0) {
			ERR("Fail to watch name [%s]", bus_name);
			free(m_info->bus_name);
			free(m_info);
			return SERVICE_COMMON_ERROR_IO_ERROR;
		}
		DBG("Watch on [%s] success", bus_name);

		monitoring_list = g_list_append(monitoring_list, strdup(bus_name));
		DBG("Success, sender[%s] length[%d]",
				sender, g_list_length(monitoring_list));
		if (g_hash_table_lookup(*monitoring_hash, GUINT_TO_POINTER(request_uid)) == NULL)
			g_hash_table_insert(*monitoring_hash, GUINT_TO_POINTER(request_uid), monitoring_list);
	} else {
		ERR("Sender [%s] already exist", sender);
	}

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		if (m_info) {
			if (m_info->bus_name)
				free(m_info->bus_name);
			free(m_info);
		}
		monitoring_list = g_list_remove(monitoring_list, bus_name);
		ERR("Failed to make reply");
		return SERVICE_COMMON_ERROR_OUT_OF_MEMORY;
	}
	return SERVICE_COMMON_ERROR_NONE;
}

int delete_monitoring_list(GHashTable **monitoring_hash, const char *sender, uid_t uid)
{
	GList *monitoring_list = NULL;
	GList *del_list = NULL;
	char *bus_name;

	monitoring_list = (GList *)g_hash_table_lookup(*monitoring_hash, GUINT_TO_POINTER(uid));
	if (monitoring_list == NULL) {
		ERR("No uid[%d] in monitoring hash", uid);
		return SERVICE_COMMON_ERROR_IO_ERROR;
	}

	monitoring_list = g_list_first(monitoring_list);
	del_list = g_list_find_custom(monitoring_list, sender,
			(GCompareFunc)_monitoring_app_list_compare_cb);

	if (del_list) {
		DBG("Find delete list - uid[%d] sender[%s]", uid, sender);
		bus_name = g_list_nth_data(del_list, 0);
		if (bus_name)
			free(bus_name);
		monitoring_list = g_list_delete_link(monitoring_list, del_list);

		if (monitoring_list == NULL) {
			g_hash_table_steal(*monitoring_hash, GUINT_TO_POINTER(uid));
		} else {
			monitoring_list = g_list_first(monitoring_list);
			g_hash_table_replace(*monitoring_hash, GUINT_TO_POINTER(uid), monitoring_list);
		}
	}
	return SERVICE_COMMON_ERROR_NONE;
}

static int _dbus_init(void)
{
	GError *error = NULL;

	if (_gdbus_conn == NULL) {
		_gdbus_conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
		if (_gdbus_conn == NULL) {
			if (error != NULL) {
				ERR("Failed to get dbus [%s]", error->message);
				g_error_free(error);
			}
			return SERVICE_COMMON_ERROR_IO_ERROR;
		}
	}

	return SERVICE_COMMON_ERROR_NONE;
}

int service_common_register_dbus_interface(char *introspection_xml, GDBusInterfaceVTable interface_vtable)
{
	int result;
	int owner_id, noti_registration_id;
	GError *error = NULL;
	GDBusNodeInfo *introspection_data = NULL;

	result = _dbus_init();
	if (result != SERVICE_COMMON_ERROR_NONE) {
			ERR("Can't init dbus [%d]", result);
			result = SERVICE_COMMON_ERROR_IO_ERROR;
			goto out;
	}

	owner_id = g_bus_own_name(G_BUS_TYPE_SYSTEM,
			PROVIDER_BUS_NAME,
			G_BUS_NAME_OWNER_FLAGS_NONE,
			NULL,
			NULL,
			NULL,
			NULL, NULL);
	if (!owner_id) {
		ERR("Failed to own name");
		result = SERVICE_COMMON_ERROR_IO_ERROR;
		goto out;
	}

	DBG("Acquiring the own name [%d]", owner_id);
	introspection_data = g_dbus_node_info_new_for_xml(introspection_xml, &error);
	if (!introspection_data) {
		ERR("g_dbus_node_info_new_for_xml is failed.");
		result = SERVICE_COMMON_ERROR_IO_ERROR;
		if (error != NULL) {
			ERR("g_dbus_node_info_new_for_xml err [%s]", error->message);
			g_error_free(error);
		}
		goto out;
	}

	noti_registration_id = g_dbus_connection_register_object(_gdbus_conn,
			PROVIDER_OBJECT_PATH, introspection_data->interfaces[0],
			&interface_vtable, NULL, NULL, NULL);

	DBG("registration id[%d]", noti_registration_id);
	if (noti_registration_id == 0) {
		ERR("Failed to g_dbus_connection_register_object");
		result = SERVICE_COMMON_ERROR_IO_ERROR;
		goto out;
	}

out:
	if (introspection_data)
		g_dbus_node_info_unref(introspection_data);

	return result;
}

static int _init_pkg_privilege_info() {

	if (_noti_pkg_privilege_info != NULL)
		return 0;

	_noti_pkg_privilege_info =
		g_hash_table_new_full(g_str_hash, g_str_equal, free, NULL);
	_badge_pkg_privilege_info =
		g_hash_table_new_full(g_str_hash, g_str_equal, free, NULL);
	DBG("init pkg privilege info done");
	return 0;
}

static int _package_install_cb(uid_t uid, const char *pkgname, enum pkgmgr_status status, double value, void *data)
{
	int ret;
	gpointer tmp;
	int privilege_info;

	if (status != PKGMGR_STATUS_END)
		return 0;

	if (uid == tzplatform_getuid(TZ_SYS_GLOBALAPP_USER))
		uid = tzplatform_getuid(TZ_SYS_DEFAULT_USER);

	_init_pkg_privilege_info();
	if (g_hash_table_contains(_noti_pkg_privilege_info, pkgname)) {
		tmp = g_hash_table_lookup(_noti_pkg_privilege_info, pkgname);
		privilege_info = GPOINTER_TO_UINT(tmp);
		if (privilege_info == 1)
			notification_setting_db_update_pkg_disabled(pkgname, false, uid);

		g_hash_table_remove(_noti_pkg_privilege_info, pkgname);
	} else {
		/* In consideration of the reboot status, change the disable information. */
		ret = notification_setting_db_update_pkg_disabled(pkgname, false, uid);
		if (ret != NOTIFICATION_ERROR_NONE)
			notification_setting_insert_package_for_uid(pkgname, uid);
	}

	if (g_hash_table_contains(_badge_pkg_privilege_info, pkgname)) {
		tmp = g_hash_table_lookup(_badge_pkg_privilege_info, pkgname);
		privilege_info = GPOINTER_TO_UINT(tmp);
		if (privilege_info == 1)
			badge_db_update_pkg_disabled(pkgname, false, uid);

		g_hash_table_remove(_badge_pkg_privilege_info, pkgname);
	} else {
		/* In consideration of the reboot status, change the disable information. */
		ret = badge_db_update_pkg_disabled(pkgname, false, uid);
		if (ret != BADGE_ERROR_NONE)
			badge_setting_insert_package_for_uid(pkgname, uid);
	}

	return 0;
}

static int _package_uninstall_cb(uid_t uid, const char *pkgname, enum pkgmgr_status status, double value, void *data)
{
	int ret;
	pkgmgrinfo_pkginfo_h pkginfo;

	if (status != PKGMGR_STATUS_END)
		return 0;

	if (uid == tzplatform_getuid(TZ_SYS_GLOBALAPP_USER))
		uid = tzplatform_getuid(TZ_SYS_DEFAULT_USER);

	ret = pkgmgrinfo_pkginfo_get_usr_disabled_pkginfo(pkgname, uid, &pkginfo);
	if (ret == PMINFO_R_OK) {
		pkgmgrinfo_pkginfo_destroy_pkginfo(pkginfo);

		_init_pkg_privilege_info();
		ret = notification_setting_db_update_pkg_disabled(pkgname, true, uid);
		if (ret == NOTIFICATION_ERROR_NONE)
			g_hash_table_insert(_noti_pkg_privilege_info, strdup(pkgname), GUINT_TO_POINTER(1));
		else
			g_hash_table_insert(_noti_pkg_privilege_info, strdup(pkgname), GUINT_TO_POINTER(0));

		ret = badge_db_update_pkg_disabled(pkgname, true, uid);
		if (ret == BADGE_ERROR_NONE)
			g_hash_table_insert(_badge_pkg_privilege_info, strdup(pkgname), GUINT_TO_POINTER(1));
		else
			g_hash_table_insert(_badge_pkg_privilege_info, strdup(pkgname), GUINT_TO_POINTER(0));
	} else {
		notification_setting_delete_package_for_uid(pkgname, uid);
		badge_db_delete_by_pkgname(pkgname, uid);
		badge_setting_delete_package_for_uid(pkgname, uid);
		notification_noti_delete_template(pkgname);
	}

	return 0;
}

static int _app_enabled_cb(uid_t uid, const char *app_id, enum pkgmgr_status status, double value, void *data)
{
	if (status == PKGMGR_STATUS_END)
		notification_setting_db_update_app_disabled(app_id, false, uid);

	return 0;
}

static int _app_disabled_cb(uid_t uid, const char *app_id, enum pkgmgr_status status, double value, void *data)
{
	if (status == PKGMGR_STATUS_END) {
		notification_delete_noti_by_app_id(app_id, uid);
		notification_setting_db_update_app_disabled(app_id, true, uid);
	}

	return 0;
}

void service_common_init(void)
{
	pkgmgr_init();
	pkgmgr_add_event_callback(PKGMGR_EVENT_INSTALL, _package_install_cb, NULL);
	pkgmgr_add_event_callback(PKGMGR_EVENT_UPDATE, _package_install_cb, NULL);
	pkgmgr_add_event_callback(PKGMGR_EVENT_UNINSTALL, _package_uninstall_cb, NULL);
	pkgmgr_add_event_callback(PKGMGR_EVENT_APP_ENABLE, _app_enabled_cb, NULL);
	pkgmgr_add_event_callback(PKGMGR_EVENT_APP_DISABLE, _app_disabled_cb, NULL);
}

void service_common_set_connection(GDBusConnection *conn)
{
	_gdbus_conn = conn;
}