/*
 * Copyright 2016  Samsung Electronics Co., Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <dlog.h>
#include <gio/gio.h>
#include <shortcut_db.h>
#include <shortcut.h>
#include "service_common.h"
#include "shortcut_service.h"
#include "debug.h"

#define PROVIDER_SHORTCUT_INTERFACE_NAME "org.tizen.data_provider_shortcut_service"
#define SHORTCUT_INVOCATION_KEY_POS 0

static GHashTable *_monitoring_hash;
static GHashTable *_invocation_hash;

static void _on_name_appeared(GDBusConnection *connection,
		const gchar     *name,
		const gchar     *name_owner,
		gpointer         user_data)
{
	DBG("name [%s]", name);
}

static void _on_name_vanished(GDBusConnection *connection,
		const gchar     *name,
		gpointer         user_data)
{
	DBG("name [%s]", name);
	monitoring_info_s *info = (monitoring_info_s *)user_data;

	if (info) {
		DBG("vanished uid [%d]", info->uid);
		g_bus_unwatch_name(info->watcher_id);
		delete_monitoring_list(&_monitoring_hash, name, info->uid);

		if (info->bus_name)
			free(info->bus_name);
		free(info);
	}
}

static void _shortcut_dbus_method_call_handler(GDBusConnection *conn,
		const gchar *sender, const gchar *object_path,
		const gchar *iface_name, const gchar *method_name,
		GVariant *parameters, GDBusMethodInvocation *invocation,
		gpointer user_data)
{
	/* TODO : sender authority(privilege) check */
	DBG("shortcut method_name [%s]", method_name);
	GVariant *reply_body = NULL;
	int ret = SHORTCUT_ERROR_NOT_SUPPORTED;
	uid_t uid = get_sender_uid(sender);

	if (g_strcmp0(method_name, "shortcut_service_register") == 0) {
		ret = service_register(parameters, &reply_body, sender,
		 _on_name_appeared, _on_name_vanished, &_monitoring_hash, uid);
	} else if (g_strcmp0(method_name, "add_shortcut") == 0) {
		shortcut_add(parameters, invocation, uid);
		return;
	} else if (g_strcmp0(method_name, "add_shortcut_widget") == 0) {
		shortcut_add_widget(parameters, invocation, uid);
		return;
	} else if (g_strcmp0(method_name, "remove_shortcut") == 0) {
		shortcut_remove(parameters, invocation, uid);
		return;
	} else if (g_strcmp0(method_name, "get_list") == 0) {
		ret = shortcut_get_shortcut_service_list(parameters, &reply_body, uid);
	} else if (g_strcmp0(method_name, "check_privilege") == 0) {
		ret = shortcut_check_privilege();
	} else if (g_strcmp0(method_name, "send_return_value") == 0) {
		ret = shortcut_send_return_value(parameters, &reply_body);
	}

	if (ret == SERVICE_COMMON_ERROR_NONE) {
		DBG("shortcut service success, method[%s]", method_name);
		g_dbus_method_invocation_return_value(
				invocation, reply_body);
	} else {
		DBG("shortcut service fail, method_name[%s] err[%d]",
			method_name, ret);
		g_dbus_method_invocation_return_error(
				invocation,
				SHORTCUT_ERROR,
				ret,
				"shortcut service error");
	}
}

static const GDBusInterfaceVTable _shortcut_interface_vtable = {
		_shortcut_dbus_method_call_handler,
		NULL,
		NULL
};

int shortcut_register_dbus_interface(void)
{
	static gchar introspection_xml[] =
			"  <node>"
			"  <interface name='org.tizen.data_provider_shortcut_service'>"
			"        <method name='shortcut_service_register'>"
			"          <arg type='i' name='uid' direction='in'/>"
			"        </method>"

			"        <method name='get_list'>"
			"          <arg type='v' name='package_name' direction='in'/>"
			"          <arg type='i' name='shortcut_cnt' direction='out'/>"
			"          <arg type='a(v)' name='shortcut_list' direction='out'/>"
			"        </method>"

			"        <method name='add_shortcut'>"
			"          <arg type='s' name='request_id' direction='in'/>"
			"          <arg type='i' name='pid' direction='in'/>"
			"          <arg type='s' name='appid' direction='in'/>"
			"          <arg type='s' name='name' direction='in'/>"
			"          <arg type='i' name='type' direction='in'/>"
			"          <arg type='s' name='uri' direction='in'/>"
			"          <arg type='s' name='icon' direction='in'/>"
			"          <arg type='i' name='allow_duplicate' direction='in'/>"
			"          <arg type='i' name='return_value' direction='out'/>"
			"        </method>"

			"        <method name='add_shortcut_widget'>"
			"          <arg type='s' name='request_id' direction='in'/>"
			"          <arg type='i' name='pid' direction='in'/>"
			"          <arg type='s' name='widget_id' direction='in'/>"
			"          <arg type='s' name='name' direction='in'/>"
			"          <arg type='i' name='size' direction='in'/>"
			"          <arg type='s' name='uri' direction='in'/>"
			"          <arg type='s' name='icon' direction='in'/>"
			"          <arg type='d' name='period' direction='in'/>"
			"          <arg type='i' name='allow_duplicate' direction='in'/>"
			"          <arg type='i' name='return_value' direction='out'/>"
			"        </method>"

			"        <method name='remove_shortcut'>"
			"          <arg type='s' name='request_id' direction='in'/>"
			"          <arg type='i' name='pid' direction='in'/>"
			"          <arg type='s' name='appid' direction='in'/>"
			"          <arg type='s' name='name' direction='in'/>"
			"          <arg type='i' name='return_value' direction='out'/>"
			"        </method>"

			"        <method name='check_privilege'>"
			"        </method>"

			"        <method name='send_return_value'>"
			"          <arg type='i' name='return_value' direction='in'/>"
			"          <arg type='s' name='request_id' direction='in'/>"
			"        </method>"
			"  </interface>"
			"  </node>";

	return service_common_register_dbus_interface(introspection_xml, _shortcut_interface_vtable);
}

static void _release_shortcut_info(gpointer data)
{
	if (data == NULL)
		return;

	shortcut_info_s *shortcut = (shortcut_info_s *)data;

	if (shortcut->package_name)
		free(shortcut->package_name);
	if (shortcut->icon)
		free(shortcut->icon);
	if (shortcut->name)
		free(shortcut->name);
	if (shortcut->extra_key)
		free(shortcut->extra_key);
	if (shortcut->extra_data)
		free(shortcut->extra_data);
	free(shortcut);
}

/* get_list */
int shortcut_get_shortcut_service_list(GVariant *parameters, GVariant **reply_body, uid_t uid)
{
	int count;
	GList *shortcut_list = NULL;
	GList *iter_list = NULL;
	GVariant *param_body = NULL;
	GVariant *body = NULL;
	shortcut_info_s *shortcut;
	GVariantDict dict;
	char *package_name = NULL;
	GVariantBuilder *builder;

	g_variant_get(parameters, "(v)", &param_body);
	g_variant_dict_init(&dict, param_body);
	g_variant_dict_lookup(&dict, "package_name", "&s", &package_name);
	g_variant_dict_end(&dict);

	count = shortcut_db_get_list(package_name, &shortcut_list);
	builder = g_variant_builder_new(G_VARIANT_TYPE("a(v)"));
	if (count > 0) {
		iter_list = g_list_first(shortcut_list);
		for (; iter_list != NULL; iter_list = iter_list->next) {
			shortcut = iter_list->data;
			body = g_variant_new("(sssss)",
				shortcut->package_name, shortcut->icon, shortcut->name, shortcut->extra_key, shortcut->extra_data);
			g_variant_builder_add(builder, "(v)", body);
		}
		g_list_free_full(shortcut_list, (GDestroyNotify)_release_shortcut_info);
	}

	*reply_body = g_variant_new("(ia(v))", count, builder);
	g_variant_builder_unref(builder);

	if (*reply_body == NULL) {
		ERR("Failed to make reply");
		return SHORTCUT_ERROR_OUT_OF_MEMORY;
	}

	INFO("count[%d]", count);
	return SERVICE_COMMON_ERROR_NONE;
}

static GDBusMethodInvocation *_get_invocation(char *request_id)
{
	GDBusMethodInvocation *find_invocation;

	find_invocation = (GDBusMethodInvocation *)g_hash_table_lookup(_invocation_hash, request_id);
	if (!find_invocation)
		return NULL;

	return find_invocation;
}

static void _add_invocation(GDBusMethodInvocation *invocation, char *request_id)
{
	GDBusMethodInvocation *find_invocation;

	if (request_id == NULL)
		return;

	find_invocation = _get_invocation(request_id);
	if (find_invocation)
		return;

	g_hash_table_insert(_invocation_hash, strdup(request_id), invocation);

	DBG("Invocation key [%s]", request_id);
}

static void _remove_invocation(char *request_id)
{
	if (request_id == NULL)
		return;

	g_hash_table_remove(_invocation_hash, request_id);

	DBG("Invocation key [%s]", request_id);
}

/* add_shortcut */
void shortcut_add(GVariant *parameters, GDBusMethodInvocation *invocation, uid_t uid)
{
	int ret = SERVICE_COMMON_ERROR_NONE;
	char *request_id = NULL;

	g_variant_get_child(parameters, SHORTCUT_INVOCATION_KEY_POS, "&s", &request_id);

	_add_invocation(invocation, request_id);

	ret = send_notify(parameters, "add_shortcut_notify", &_monitoring_hash, PROVIDER_SHORTCUT_INTERFACE_NAME, uid);
	if (ret != SERVICE_COMMON_ERROR_NONE)
		ERR("Failed to send notify [%d]", ret);
	else
		DBG("Success to send notify");

}

/* add_shortcut_widget */
void shortcut_add_widget(GVariant *parameters, GDBusMethodInvocation *invocation, uid_t uid)
{
	int ret = SERVICE_COMMON_ERROR_NONE;
	char *request_id = NULL;

	g_variant_get_child(parameters, SHORTCUT_INVOCATION_KEY_POS, "&s", &request_id);

	_add_invocation(invocation, request_id);

	ret = send_notify(parameters, "add_shortcut_widget_notify", &_monitoring_hash, PROVIDER_SHORTCUT_INTERFACE_NAME, uid);
	if (ret != SERVICE_COMMON_ERROR_NONE)
		ERR("Failed to send notify [%d]", ret);
	else
		DBG("Success to send notify");
}

/* remove_shortcut */
void shortcut_remove(GVariant *parameters, GDBusMethodInvocation *invocation, uid_t uid)
{
	int ret = SERVICE_COMMON_ERROR_NONE;
	char *request_id = NULL;

	g_variant_get_child(parameters, SHORTCUT_INVOCATION_KEY_POS, "&s", &request_id);

	_add_invocation(invocation, request_id);

	ret = send_notify(parameters, "remove_shortcut_notify", &_monitoring_hash, PROVIDER_SHORTCUT_INTERFACE_NAME, uid);
	if (ret != SERVICE_COMMON_ERROR_NONE)
		ERR("Failed to send notify [%d]", ret);
	else
		DBG("Success to send notify");
}

/*  check shortcut privilege */
int shortcut_check_privilege(void)
{
	return SERVICE_COMMON_ERROR_NONE;
}

int shortcut_send_return_value(GVariant *parameters, GVariant **reply_body)
{
	int return_value;
	char *request_id = NULL;
	GDBusMethodInvocation *invocation = NULL;

	g_variant_get(parameters, "(i&s)", &return_value, &request_id);

	if (request_id == NULL)
		return SHORTCUT_ERROR_INVALID_PARAMETER;

	invocation = _get_invocation(request_id);
	if (!invocation)
		return SHORTCUT_ERROR_INVALID_PARAMETER;

	g_dbus_method_invocation_return_value(invocation, g_variant_new("(i)", return_value));
	_remove_invocation(request_id);

	*reply_body = g_variant_new("()");
	if (*reply_body == NULL) {
		ERR("Failed to make reply");
		return SHORTCUT_ERROR_OUT_OF_MEMORY;
	}

	DBG("Success to send return value");
	return SHORTCUT_ERROR_NONE;
}

/*!
 * MAIN THREAD
 * Do not try to do anyother operation in these functions
 */
HAPI int shortcut_service_init(void)
{
	int result;

	_monitoring_hash = g_hash_table_new(g_direct_hash, g_direct_equal);
	_invocation_hash = g_hash_table_new_full(g_str_hash, g_str_equal, free, NULL);

	result = shortcut_db_init();
	if (result != SHORTCUT_ERROR_NONE) {
		ERR("Failed to init DB[%d]", result);
		return result;
	}

	result = shortcut_register_dbus_interface();
	if (result != SERVICE_COMMON_ERROR_NONE) {
		ERR("Failed to register dbus interface [%d]", result);
		return result;
	}

	INFO("Initialization success");
	return result;
}

HAPI int shortcut_service_fini(void)
{
	g_hash_table_destroy(_invocation_hash);

	DBG("Finalization success");
	return SERVICE_COMMON_ERROR_NONE;
}

/* End of a file */
